#ifndef MT_algos_h
#define MT_algos_h

#include "array.h"
#include "list.h"
//#include "math.h"
//#include "regression.h"


namespace MT{

  /*! Singular Value Decomposition (from Numerical Recipes); computes
    $U, W, V$ with $A = UWV^T$ from $A$ such that $U$ and $V$ are
    orthogonal and $W$ diagonal (the returned array #w# is
    1-dimensional)*/
  uint svd(const doubleA& A,doubleA& U,doubleA& V,doubleA& w,bool sort=false);

  inline void eigensymm(const doubleA& A,doubleA& V,doubleA& w){
    doubleA U;
    svd(A,U,V,w,true);
  }


  /*! Pseudo Inverse based on SVD; computes $B$ such that $ABA = A$ */
  uint inverse(const doubleA& A,doubleA& inverse);

  /*! the determinant of a 2D squared matrix */
  double determinant(const doubleA& A);

  /*! the cofactor is the determinant of a 2D squared matrix after removing
      the ith row and the jth column */
  double cofactor(const doubleA& A,uint i,uint j);

  /*! stochastic universal sampling: given a probability vector #p#
    (that is positive in each component and sums to one), the
    returned array #s# is resized to length #n# and contains the
    indices of #n# samples. */
  void SUS(const doubleA& p,uint n,Array<uint>& s);

  /*! as above, but returns only the index of one sample */
  uint SUS(const doubleA& p);

  extern Parameter <uint> _annealingRepetitions_;
  extern Parameter <double> _annealingCooling_;

  /*! the Hamming distance between two matrices #fix# and #fox# when
      the nodes of #fox# are permuted according to #p#. If they don't
      match in size and #sub=1#, then only the upper left submatrices
      are compared; if #sub=0#, the missing entries of the smaller
      matrix are counted as wrong symbols */
  template<class X>
  double matdistance(Array<X>& fix,Array<X>& fox,Permutation& p,bool sub){
    CHECK(fox.d0==p.N,"matrix and its permutation don't agree in size");
    uint i,j,Nmin,Nmax,n=0;
    if(fix.d0<=fox.d0){ Nmin=fix.d0; Nmax=fox.d0; }
    else{ Nmin=fox.d0; Nmax=fix.d0; }
    for(i=0;i<Nmin;i++) for(j=0;j<Nmin;j++) if(fix(i,j)!=fox(p(i),p(j))) n++;
    if(!sub) n+=(Nmax-Nmin)*(Nmax+Nmin);
    //std::cout <<fix <<fox <<n;
    return ((double)n)/Nmax/Nmax;
  }

  /*! same as #matdistance(A,B,identical-permutation,sub)# */
  template<class X>
  double matdistance(Array<X>& A,Array<X>& B,bool sub){
    Permutation p;
    p.init(B.d0);
    return matdistance(A,B,p,sub);
  }

  /* Uses simulated annealing to find the permutation #p# such that
     #matdistance(fix,fox,p,sub)# becomes minimal */
  template<class X>
  double matannealing(Array<X>& fix,Array<X>& fox,Permutation& p,bool sub){
    CHECKO(fix.nd==2 && fox.nd==2);

    uint N=fox.d0;
    p.init(N); p.store();
    if(fox.N<=1 && fix.N<1) return matdistance(fix,fox,p,sub);
  
    uint i,j,k;
    unsigned long t;
    double newdist,dist,bestdist=1,temp;
  
    for(k=0;k<_annealingRepetitions_;k++){
      t=N*N/10;
      p.random();
      temp=0.1;
      dist=matdistance(fix,fox,p,sub);
      if(dist<bestdist){ if(dist==0) return 0; bestdist=dist; p.store(); }
      while(temp>0.001){
	i=rnd(N); j=rnd(N-1); if(j>=i) j++;
	p.permute(i,j);
	newdist=matdistance(fix,fox,p,sub);
	if(::exp(-(newdist-dist)/temp)>rnd.uni()){
	  dist=newdist;
	  if(dist<bestdist){
	    if(dist==0) return 0;
	    bestdist=dist;
	    p.store();
	  }
	}else{
	  p.permute(i,j);
	}
	if(!(t--)){ t=N*N/10; temp*=_annealingCooling_; }
      }
    }
    p.restore();
    return bestdist;
  }
}


class MonSolver{
public:
  double min,max;
  int phase;
  MonSolver(){
    min=-.1; max=.1;
    phase=0;
  }
  void init(double& par,double wide=2.){
    phase=0;
    double w=.5*wide*(max-min);
    max+=w;
    min-=w;
    par=min;
  }
  void solve(double& par,const double& err){
    if(phase==0){
      CHECK(par==min,"not phase 0!");
      if(err>0.){
	min-=max-min;
	par=min;
	return;
      }else{
	phase=1;
	par=max;
	return;
      }
    }
    if(phase==1){
      CHECK(par==max,"not phase 1!");
      if(err<0.){
	max+=max-min;
	par=max;
	return;
      }else{
	phase=2;
	par=.5*(max+min);
	return;
      }
    }
    if(phase==2){
      if(err>0.) max=par; else min=par;
      par=.5*(max+min);
    }
  }
};

//===========================================================================
//
// Linear Statistics
//

/*! collects means and (co-)variances of data, also input-output data.
    This class is the basis for linear regression techniques, like Partial
    Least Squares, but can also be used on its own, simply to average or calculate
    (co-)variances \ingroup regression */
class LinearStatistics{
public:
  double accum;  //!< the accumulation norm (=number of collected data points if not weighted)
  doubleA meanX,meanY,varX,covXY; //<! these are not normalized or centered bufferes
  doubleA MeanX; //!< X mean
  doubleA MeanY; //!< Y mean
  doubleA VarX;  //!< X variance
  doubleA CovXY; //!< XY covariance
  Parameter<double> lambda; //!< forgetting rate [default=0]

  bool computed; //!< internal indicator whether recomputation is needed

  LinearStatistics():
  lambda("LinearStatisticsForgettingRate",0.){ accum=0.; computed=false; }

  /*! NOTE: when data is added to the statistics, the means and
  variances are not yet normalized and centered (computed indicates that) */
  void compute();

  void computeZeroMean();

  /*! add a single new datum (x and y are 1D arrays)
      or a batch of data (x and y are 2D arrays) to the statistics in a weighed fashion */
  void learn(const doubleA& x,const doubleA& y,double weight);

  //! in case y is a scalar
  void learn(const doubleA& x,double y,double weight){ learn(x,doubleA(&y,1),weight); }

  //! in case there is no y
  void learn(const doubleA& x){ if(x.nd==1) learn(x,doubleA(0),1.); else learn(x,doubleA(x.d0,0),1.); }

  //! weighted learning
  void learn(const doubleA& x,double weight){ if(x.nd==1) learn(x,doubleA(0),weight); else learn(x,doubleA(x.d0,0),weight); }

  void clear(){ forget(1); }

  //! forget the current statistics partially
  void forget(double lambda=1.);

  uint inDim(){ return meanX.N; }

  double variance(){ compute(); return trace(VarX); }

  void correlation(doubleA& corr,bool clearDiag=false){
    compute();
    corr=VarX;
    doubleA sd(corr.d0);
    uint i,j;
    for(i=0;i<corr.d0;i++) sd(i)=::sqrt(corr(i,i));
    for(i=0;i<corr.d0;i++) for(j=0;j<corr.d1;j++)
      if(sd(i)*sd(j)) corr(i,j)/=sd(i)*sd(j); else corr(i,j)=0.;
    if(clearDiag) for(i=0;i<corr.d0;i++) corr(i,i)=0.;
  }

  //! returns the mahalanobis metric relative to the VarX matrix
  void mahalanobisMetric(doubleA& metric){
    compute();
    metric.resize(VarX);
    doubleA var;
    var=VarX;
    for(uint i=0;i<var.d0;i++) var(i,i)+=1./(1.+accum);
    MT::inverse(metric,var);
  }

  //! prototype for operator<<
  void write(std::ostream& os) const{
    os
      <<"<LinearStatistics>"
      <<"\ndim X=" <<MeanX.N
      <<", dim Y=" <<MeanY.N
      <<", accum=" <<accum
      <<"\nmean X=" <<MeanX.ioraw()
      <<"\nmean Y=" <<MeanY
      <<"\nvariance X="<<VarX
      <<"\ncovariance XY=" <<CovXY
      <<"\n</LinearStatistics>"
      <<std::endl;
  }
};
stdOutPipe(LinearStatistics);

class Rprop{
public:
  double delta0;
  double np;
  double nm;
  double dMax;
  double dMin;
  double rMax;
  
  doubleA deltaw;     
  doubleA deOld; // last error gradient
  doubleA delta; // last update

  Rprop(){
    np   = 1.2;
    nm   = .5;
    dMax = 50;
    dMin = 1e-6;
    rMax = 0;
  }

  void init(uint size,double _delta0 = 1.) { 
    deltaw.resize(size);
    delta.resize (size);
    deOld.resize (size);
    delta0 = _delta0;
    delta  = delta0; 
    deltaw = 0; 
    deOld  = 0; 
  }
  void forget(){ init(delta.N); }
  void learn(doubleA& w,const doubleA& dedw){
    if(!delta.N) init(dedw.N);
    CHECK(dedw.N==delta.N,"gradient dimensionality changed!");

    for(unsigned  i = 0; i < w.N; i++) {
      if(dedw.elem(i) * deOld(i) > 0){
	if(rMax) dMax=fabs(rMax*w.elem(i));
	delta(i) = min(dMax, np * delta(i));
	deltaw(i) = delta(i) * -sgn(dedw.elem(i));
	w.elem(i) += deltaw(i);
	deOld(i) = dedw.elem(i);
      }else if(dedw.elem(i) * deOld(i) < 0) {
	delta(i) = max(dMin, nm * delta(i));
	deOld(i) = 0;
      }else{
	deltaw(i) = delta(i) * -sgn(dedw.elem(i));
	w.elem(i) += deltaw(i);
	deOld(i) = dedw.elem(i);
      }
    }
  }
  void learn(double& w,const double& dedw){
    static doubleA W(1),DEDW(1);
    W(0)=w; DEDW(0)=dedw; learn(W,DEDW); w=W(0);
  }

private:
  int sgn(double x){ if (x > 0) return 1; if (x < 0) return -1; return 0; }
  double min(double x,double y){ return x < y ? x : y; }
  double max(double x,double y){ return x > y ? x : y; }
};

class TupleIndex:public uintA{
public:
  uintA tri;

  //! initializes to ``k out of n''
  void init(uint k,uint n){
    CHECK(k<=n,"TupleIndex: can't choose "<<k<<" out of "<<n);
    uint i,j;
    tri.resize(n+1,k+1); tri=0;
    for(i=0;i<=n;i++) for(j=0;j<=i && j<=k;j++){
      if(i==0 || j==0 || j==i){
        tri(i,j)=1;
      }else{
	tri(i,j)=tri(i-1,j-1) + tri(i-1,j);
      }
    }

    uint N=tri(n,k);
    resize(N,k);
    for(j=0;j<k;j++) operator()(0,j)=j;
    for(i=1;i<N;i++){
      for(j=0;j<k;j++) operator()(i,j)=operator()(i-1,j);
      for(j--;operator()(i,j)==n+j-k;j--);
      operator()(i,j)++;
      for(j++;j<k;j++) operator()(i,j)=operator()(i,j-1)+1;
    }
  }

  uint index(uintA i){
    uint a=0,n=tri.d0-1,k=i.N,u;
    if(k) a+=tri(n,k)-tri(n-i(0),k);
    for(u=1;u<i.N;u++){
      CHECK(i(u)>i(u-1) && i(u) < n,"wrong symmetic index!");
      a+=tri(n-i(u-1)-1,k-u)-tri(n-i(u),k-u);
    }
    std::cout <<a <<std::endl;
    return a;
  }

  void checkValid(){
    uint i;
    for(i=0;i<d0;i++) CHECK(i==index(operator[](i)),"wrong index association");
  }
};





namespace MT{
  void normalizeData(doubleA& X);
}



//---------------------------------------------------------------------------
//
// X-Splines
//

class XSpline
{
public:
  double DELTA;     // Distance between each

  //double inc;       // change sk in inc increments
  //double *sv;       // Weight doubleA
  //doubleA *cv;       // Control doubleA
  doubleA V; //vertices
  doubleA W; //weights associated to each vertex

public:

  XSpline(){ DELTA = 1.0; };
  ~XSpline(){};
   
  void setWeights(double w){
    W.resize(V.d0); W=w; W(0)=W(W.N-1)=0.;
  }
  doubleA eval(double t);
  //const doubleA& get_cp(int k) const {return V[k];}
  //void set_cp(int k, const doubleA& pos) {V[k]=pos;}
  //void inc_sk(int k) {if (sv[k]<1.0-inc) sv[k] += inc;}
  //void dec_sk(int k) {if (sv[k]>-1.0+inc) sv[k] -= inc;}
  // Hopefully the function below is faster than the 
// generic pow function for integer powers.
private:
double intpow(const double x, const int n ){
  double val = x;
  for (int i=1;i<n;i++)
    val *= x;
  return val;
}


// g(u) as defined in the above mentioned article
double g(double u, double q, double p){
  return 
    q *                                    u 
      + 2 * q *                     intpow(u,2)
	+ (10 - 12 * q - p) *       intpow(u,3)
	  + (2 * p + 14 * q - 15) * intpow(u,4)
	    + (6 - 5*q - p) *       intpow(u,5);
}


// h(u)
double h(double u, double q){
  return 
    q *                    u
      + 2 * q *     intpow(u,2)  
	- 2 * q *   intpow(u,4)
	  - q *     intpow(u,5);
}

};



//---------------------------------------------------------------------------
//
// Approximate Nearest Neighbor Search (kd-trees, etc)
//

#ifdef MT_ANN
#include<ANN/ANN.h>

struct doublePair{ doubleA x,y; };
std::ostream& operator<<(std::ostream& os,const doublePair& p){ os <<p.x <<p.y; return os; }

class ANN{
public:
  MT::List<doublePair> X;
  MT::Array<double*> C;
  intA L;
  uint D,Dy,treeS;

private:
  ANNbd_tree *tree;
  PartialLeastSquares pls;

public:

  ANN(){ tree=0; treeS=0; D=Dy=0; }
  ~ANN(){ if(tree) delete tree; annClose(); }

  void learn(const doubleA& x,const doubleA& y){ add(x,&y); }

  void add(const doubleA& x,const doubleA* y=0){
    if(x.nd==2){ for(uint i=0;i<x.d0;i++) add(x[i]); return; }
    if(!D) D=x.N;
    if(y && !Dy) Dy=y->N;
    CHECK(D==x.N,"");
    CHECK((!y && Dy==0) || (y && Dy==y->N),"");

    //find closest so far
    int l;
    double d;
    if(tree){
      tree->annkSearch(x.p,1,&l,&d,0.);
      if(d<1e-20){ MT_MSG("already added! - yet not handled case!"); return; }
    }

    //add to data
    doublePair p; p.x.referTo(x); if(y) p.y.referTo(*y);
    X.append(p);
    L.append(-1);
    if(tree){
      if(true || X.N-treeS > .1*X.N){
	L=-1;
	delete tree;
	tree=0;
      }else{
	while(L(l)!=-1) l=L(l);
	L(l)=X.N-1;
      }
    }
  }

  void calculate(){
    if(tree) return;
    C.resize(X.N);
    for(uint i=0;i<X.N;i++) C(i)=X(i)->x.p;
    tree = new ANNbd_tree(C.p,X.N,D);
    treeS=X.N;
  }

  void getNN(const doubleA& x,uint k,intA& idx,double eps=.0,bool talk=false){
    CHECK(X.N>=k,"data has less ("<<X.N<<") than k="<<k<<" points");
    CHECK(x.N==D,"query point has wrong dimension");
    calculate();
    doubleA dists(k);
    idx.resize(k);
    tree->annkSearch(x.p,k,idx.p,dists.p,eps);
    uint i;
    int l;
    for(i=0;i<k;i++){
      l=L(idx(i));
      while(l!=-1){
	idx.append(l);
	dists.append(sqrDistance(x,X(l)->x));
	l=L(l);
      }
    }
    if(talk){
      std::cout
	<<"ANN query:\n query point "<<x
	<<"\n found neighbors:\n";
      for(uint i=0;i<idx.N;i++){
	std::cout <<' '
	  <<i <<' '
	  <<idx(i) <<'\t'
	  <<sqrt(dists(i)) <<'\t'
	  <<X(idx(i))->x <<std::endl;
      }
    }
  }

  void getNN(doubleA& x,uint k,doubleA& xx,double eps=.0,bool talk=false){
    intA idx;
    getNN(x,k,idx,eps,talk);
    xx.resize(idx.N,D);
    for(uint i=0;i<idx.N;i++) xx[i]=X(idx(i))->x;
  }

  void map(const doubleA& x,doubleA& y){
    if(x.nd==2){ y.resize(x.d0,Dy); for(uint k=0;k<x.d0;k++) map(x[k],y[k]()); return; }
    if(!X.N) return;
    calculate();
    uint nn=20;
    if(nn>treeS) nn=treeS;
    doubleA dists(nn);
    intA idx(nn);
    tree->annkSearch(x.p,nn,idx.p,dists.p,0.);
    LinearStatistics S;
    uint i;
    for(i=0;i<nn;i++) S.learn(X(idx(i))->x-x);
    //S.computeZeroMean();
    double d=.5*S.variance(),w;
    if(!d) d=1.;
    pls.S.forget();
    S.forget();
    for(i=0;i<nn;i++){
      w=::exp(-.5*dists(i)/d);
      w=1./(nn+i);
      S.learn(X(idx(i))->y,w);
      //pls.learn(X(idx(i))->x,X(idx(i))->y,w);
    }
    S.compute();
    y=S.MeanX;
    //pls.map(x,y);
  }
};
#endif


#ifdef MT_IMPLEMENTATION
#  include"algos.cpp"
namespace MT{
  Parameter <uint> _annealingRepetitions_ ("annealingRepetitions",100);
  Parameter <double> _annealingCooling_ ("annealingCooling",.9);
}
#endif

#endif
