#ifndef _codeMap_h_
#define _codeMap_h_

#ifndef MT_MSVC
#  include <ext/hash_set>
   using __gnu_cxx::hash_set;
#else
#  include <hash_set>
   using std::hash_set;
   /*MSVC NOTE: To get portable hash_set support,I use the stlport
     (version 4.6.2), `prepared' for vc6, without own IOSTREAM builds */
#endif
#include "array.h"
#include "list.h"

/*! a production replaces symbols (LHS) in a string by a substring (RHS) */
struct Production{
public:
  int
    code; //!< the code (LHS, promoter) of this production
  MT::Array<int>
    str;  //!< the RHS (structural genes) of this production
  uint
    c,    //!< counter of occurencies
    h,    //!< counter of hierarchy
    size; //!< total size, also accounting for the size the symbols represent
  Production(){ code=-1; c=h=size=0; }
};
struct codehashfun{
  size_t operator()(const Production& a) const{ return a.str(0); }
  bool operator()(const Production& a,const Production& b) const{ return a.str==b.str; }
};
struct decodehashfun{
  size_t operator()(const Production& a) const{ return a.code; }
  bool operator()(const Production& a,const Production& b) const{ return a.code==b.code; }
};
typedef hash_set<Production,codehashfun,codehashfun> codehash;
typedef hash_set<Production,decodehashfun,decodehashfun> decodehash;

struct SortOp{
  //lower operator
  bool operator()(MT::List<Production>::node a,MT::List<Production>::node b){
    return compare(a->data(),b->data())==1; }

  static int compare(const Production &o,const Production &u){
    //favor higher frequencies
    if(o.c>u.c) return 1;
    if(o.c<u.c) return -1;
#if 0
    if(!onlyFrequencies){
      //favor lower hierarchies
      if(o.h>u.h) return 1;
      if(o.h<u.h) return -1;
      //favor larger overall sizes
      if(o.size>u.size) return 1;
      if(o.size<u.size) return -1;
    }
#endif
    return 0;
  }
};


/*! the CodeMap computes a production system (L-system) to compress/reprepresent
    a set (population) of strings */
class CodeMap{
public:
  codehash coding;
  decodehash decoding;
  codehash pairs;
  int maxcode;
  Parameter<bool> forceSameHierarchy;

  CodeMap():
    forceSameHierarchy("CodeMapForceSameHierarchy"){
    maxcode=0;
  }

  void clear(){ coding.clear(); decoding.clear(); pairs.clear(); }

private:

public:
  /*! searches for the best (most occuring) symbol pair in the pop
      to be replaced and add it to the code & decode map */
  template<class Pop>
  bool addBestPairToMap(Pop& pop,bool onlyBest=true,int hi=0){
    codehash::iterator i;
    typename Pop::node s; uint j;
    Production r; r.str.resize(2);
    pairs.clear();

    if(hi) maxcode=hi;
    else{
      maxcode=0;
      for(i=coding.begin();i!=coding.end();i++) if((*i).code>maxcode) maxcode=(*i).code;
    }

    //collect a set (pairs) of possible Production's and count how often they occur
    forall(s,pop){
      for(j=1;j<s->N;j++){
	//construct a possible Production
        r.str(0)=s->elem(j-1); r.str(1)=s->elem(j);
	//see if the pairs is already in the code map
        i=coding.find(r);
	if(i==coding.end()){
  	  //see if its pairs occured before
	  i=pairs.find(r);
	  if(i!=pairs.end()){ //if yes, increase the Production's frequency counter
	    ((Production&)(*i)).c++;
	  }else{ //if no, insert it in the set (pairs) of previously occured pairs
	    //if(checkProduction(r)){
	      r.c=1;
	      pairs.insert(r);
	    //}
	  }
        }
	//keep also track what the maximal symbol value is in all pop:
        if(s->elem(j)>maxcode) maxcode=s->elem(j);
      }
      if(s->N && s->elem(0)>maxcode) maxcode=s->elem(0);
    }

#if 1
    //copy pairs into list:
    MT::List<Production> list;
#ifndef ALSO_ONE_PAIRS
    for(i=pairs.begin();i!=pairs.end();i++) if((*i).c>1) list.add(*i);
#else
    for(i=pairs.begin();i!=pairs.end();i++) list.add(*i);
#endif
    if(!list.N) return false; //return false, if no Productions were found

    //sort list according to compare operator
    list.randomIndexing();
    list.sortIndexing(SortOp());
    //if(list(0)->c<2) return false; //return false if the best Production has length <2

    if(onlyBest){
      //add only best to the code&decode maps
      maxcode++; //take the next largest code as the Productions code
      list(0)->code=maxcode;
      coding.insert(*list(0));
      decoding.insert(*list(0));
    }else{
      for(j=0;j<list.N;j++){
        maxcode++; //take the next largest code as the Productions code
        list(j)->code=maxcode;
        coding.insert(*list(j));
	decoding.insert(*list(j));
      }
    }
    return true;
#else//older code
    //pick the production that compares better to all others ('compare' routine!)
    const Production *max=0;
    //first collect a list of all productions that compare equally
    MT::Array<Production> maxlist;
    for(i=pairs.begin();i!=pairs.end();i++){
      if(!max || (*i).c>(*max).c){
        max=&(*i);
        maxlist.resize(0);
        maxlist.append(*i);
      }else if((*i).c>(*max).c) maxlist.append(*i);
    }

    //then select randomly one production out of them and add it to the code&decode maps
    if(max && max->c>1){ //return false if the best Production has length <2
      max=&maxlist(rnd(maxlist.N));
      maxcode++; //take the next largest code as the Production's code
      ((Production&)*max).code=maxcode;
      coding.insert(*max);
      decoding.insert(*max);
      return true;
    }else return false;
#endif
  }

  /*! in all the pop, search for pairs that can be recoded
      by an Production in the code map and replace them by the Production's code */
  template<class Pop>
  bool recode(Pop& pop){
    codehash::iterator i;
    typename Pop::node s; uint j;
    bool mod=false;
    Production r; r.str.resize(2);
    forall(s,pop) for(j=1;j<s->N;j++){
      r.str(0)=s->elem(j-1); r.str(1)=s->elem(j);
      i=coding.find(r);
      if(i!=coding.end()){
        s->remove(j-1);
        s->elem(j-1)=(*i).code;
        j--;
	mod=true;
      }
    }
    return mod;
  }

  template<class Pop>
  void model(Pop& pop){ HALT("can't just model a population"); }
  
  //! iterates addBestPairToMap(pop) and recode(pop)
  template<class Pop>
  void encode(Pop& pop,double abort=0.){
    clear();
    maxcode=0;
    uint i;
    for(i=0;;i++){
      //while(addBestPairToMap(pop));
      if(abort){
	if(abort<1. && rnd.uni()<abort) break;
        if(abort>=1. && i>=abort) break;
      }
      addBestPairToMap(pop,false);
      if(!recode(pop)) break;
    }
  }

  /*! as above, but calls a mutation operator
      MutateStruct::mutate(Pop& s,int lowestSymbol,int highestSymbols)
      for each sample on each level of encoding */
  template<class Pop,class MutationOp>
  void encodeAndMute(Pop& pop,const MutationOp&,int lo=0,int hi=0){
    clear();
    typename Pop::node s;
    for(;;){
      //while(addBestPairToMap(pop));
      addBestPairToMap(pop,false,hi);
      if(!recode(pop)) break;
      forall(s,pop) MutationOp::mutate(*s,lo,maxcode,.1);
    }
  }


private:
  void calcSizes(){
    decodehash::iterator i;
    for(i=decoding.begin();i!=decoding.end();i++) checkProduction((Production&)*i);
  }
  bool checkProduction(Production& op){
    decodehash::iterator i;
    Production o;
    uint j;
    bool sameH=true;
    //check for the first symbol
    o.code=op.str(0);
    i=decoding.find(o);
    if(i==decoding.end()){ op.size=1; op.h=1; }else{ op.size=(*i).size; op.h=(*i).h+1; }
    //check for the rest of symbols
    for(j=1;j<op.str.N;j++){
      o.code=op.str(j);
      i=decoding.find(o);
      if(i==decoding.end()){ op.size+=1; if(1!=op.h) sameH=false; }else{ op.size+=(*i).size; if((*i).h+1!=op.h) sameH=false; }
    }
    if(forceSameHierarchy) return sameH;
    return true;
  }

  void expand(MT::Array<int>& s){
    decodehash::iterator i;
    uint j;
    Production op;
    for(j=0;j<s.N;j++){
      op.code=s(j);
      i=decoding.find(op);
      if(i!=decoding.end()){ s.replace(j,1,(*i).str); j--; }
    }
  }


public:
  /*! decodes all pop by replacing all codes (LHS)
      of the productions in the decode map by their RHS */
  template<class Pop>
  void decode(Pop& pop){
    decodehash::iterator i;
    typename Pop::node s; uint j;
    Production op;
    forall(s,pop) for(j=0;j<s->N;j++){
      op.code=s->elem(j);
      i=decoding.find(op);
      if(i!=decoding.end()){
        s->replace(j,1,(*i).str);
        j--;
      }
    }
  }


  /*! expands all RHSs of the productions, e.g., to look at them */
  void expandMap(){
    decodehash::iterator i;
    for(i=decoding.begin();i!=decoding.end();i++) expand(((Production&)(*i)).str);
  }

  static void writeCode(std::ostream& os,int i){
    if(i>=52) os <<'.' <<i; else{
      if(i<26) os <<(char)('A'+i); else os <<(char)('a'+i-26); } }
  static void writeSeq(std::ostream& os,const MT::Array<int>& s){
    for(uint i=0;i<s.N;i++) writeCode(os,s(i)); }
  static void writeOp(std::ostream& os,const Production& o){
    writeCode(os,o.code);
    os <<'-';
    writeSeq(os,o.str);
    os <<':' <<o.c <<':' <<o.size <<':' <<o.h <<std::endl;
  }

  //! operator<<
  void write(std::ostream& os) const{
    codehash::iterator i;
    decodehash::iterator j;
#if 1
    os <<"pairs: " <<pairs.size() <<std::endl;
    for(i=pairs.begin();i!=pairs.end();i++) writeOp(os,*i);
#endif
    os <<"coding: " <<coding.size() <<std::endl;
    for(i=coding.begin();i!=coding.end();i++) writeOp(os,*i);
    os <<"decoding: " <<decoding.size() <<std::endl;
    for(j=decoding.begin();j!=decoding.end();j++) writeOp(os,*j);
  }
};
stdOutPipe(CodeMap);


#endif
