/*! \file distributions.h */
#ifndef MT_distributions_h
#define MT_distributions_h

#include "list.h"
#include "graph.h"
#include "algos.h"


//===========================================================================
//
// analyzing samples
//

namespace MT{
  //! the maximal length of all sequences
  uint maxLen(const intAL& POP){
    uint l=0;
    intAL::node n;
    forNodes(n,POP) if(n->N>l) l=n->N;
    return l;
  }

  //! the maximal symbol in all sequences
  uint maxSym(const intAL& POP){
    uint s=0,i;
    intAL::node n;
    forNodes(n,POP) for(i=0;i<n->N;i++) if(n->operator()(i)>(int)s) s=n->operator()(i);
    return s;
  }

  /*! the length distributions */
  void lengthDist(const intAL& POP,doubleA& l,uint maxLength){
    l.resize(maxLength+1); l.setZero();
    double Z=0.;
    intAL::node n;
    forNodes(n,POP){ l(n->N)+=n->w; Z+=n->w; }
    if(Z) l/=Z; else l=1./l.N;
  }

  /*! 1D array: probabilities $p(X_i=x)$ for given i */
  void marginal(const intAL& POP,doubleA& p,uint i,int maxSymbol){
    p.resize(maxSymbol+1); p.setZero();
    intAL::node n;
    double Z=0.;
    int xi;
    forNodes(n,POP) if(i<n->N){
      xi=n->p[i];
      if(xi!=-1){ p(xi)+=n->w; Z+=n->w; }
    }
    if(Z) p/=Z; else p=1./p.N;
  }
  /*! 2D array: probabilities $p(X_i=x)$ for all i */
  void marginals(const intAL& POP,doubleA& p,uint maxLength,int maxSymbol){
    p.resize(maxLength,maxSymbol+1); p.setZero();
    for(uint i=0;i<maxLength;i++) marginal(POP,p[i](),i,maxSymbol);
  }

  /*! 1D array of 1D array's: probabilities $p(X_i=x)$ for all i */
  void marginals(const intAL& POP,MT::Array<doubleA>& p,uint maxLength,int maxSymbol){
    p.resize(maxLength);
    for(uint i=0;i<maxLength;i++){
      p(i).resize(maxSymbol+1); p(i).setZero();
      marginal(POP,p(i),i,maxSymbol);
    }
  }

  /*! 2D array: probabilities $p(X_i=x,X_j=y)$ */
  void pair(const intAL& POP,doubleA& p,uint i,uint j,int maxSymbol){
    p.resize(maxSymbol+1,maxSymbol+1); p.setZero();
    intAL::node n;
    double Z=0.;
    uint nN; int xi,xj;
    forNodes(n,POP){
      nN=n->N;
      if(i<nN && j<nN){
	xi=n->p[i];
	xj=n->p[j];
	if(xi!=-1 && xj!=-1){ p(xi,xj)+=n->w; Z+=n->w; }
      }
    }
    if(Z) p/=Z; else p=1./p.N;
  }

  /*! 2D array: conditional probabilities $p(X_i=x,X_j=y|X_k=c)$ */
  void pair(const intAL& POP,doubleA& p,uint i,uint j,int maxSymbol,uint k,int c){
    p.resize(maxSymbol+1,maxSymbol+1); p.setZero();
    intAL::node n;
    double Z=0.;
    forNodes(n,POP) if(i<n->N && j<n->N && k<n->N
		    && n->operator()(i)!=-1 && n->operator()(j)!=-1
		    && n->operator()(k)==c){ p(n->operator()(i),n->operator()(j))+=n->w; Z+=n->w; }
    if(Z) p/=Z; else p=1./p.N;
  }

  /*! 3D array: probabilities $p(X_i=x,X_j=y,X_k=z)$ */
  void triple(const intAL& POP,doubleA& p,uint i,uint j,uint k,int maxSymbol){
    p.resize(maxSymbol+1,maxSymbol+1,maxSymbol+1); p.setZero();
    intAL::node n;
    double Z=0.;
    uint nN; int xi,xj,xk;
    forNodes(n,POP){
      nN=n->N;
      if(i<nN && j<nN && k<nN){
	xi=n->p[i];
	xj=n->p[j];
	xk=n->p[k];
	if(xi!=-1 && xj!=-1 && xk!=-1){ p(xi,xj,xk)+=n->w; Z+=n->w; }
      }
    }
    if(Z) p/=Z; else p=1./p.N;
  }

  /*! 1D array: probabilities $p(X_{\{{\rm loci}\}}=x$; indexing is done by
      n-ary numbering, where n is the alphabet size */
  void tuple(const intAL& POP,doubleA& p,const uintA& loci,int maxSymbol){
    uint i,j,k=loci.N;
    p.resize((uint)::pow(maxSymbol+1,k)); p.setZero();
    intAL::node n;
    double Z=0.;
    bool discard;
    forNodes(n,POP){
      discard=false;
      for(i=0;i<k;i++) if(!(loci(i)<n->N) || n->operator()(loci(i))==-1) discard=true;
      if(!discard){
	j=0;
	for(i=0;i<k;i++){ j*=maxSymbol+1; j+=n->operator()(loci(i)); }
	p(j)+=n->w;
	Z+=n->w;
      }
    }
    if(Z) p/=Z; else p=1./p.N;
  }

  /*! entropy when assuming that all sequences are different
      (only based on the weights) */
  double entropy(const intAL& POP){
    intAL::node n;
    double H=0.;
    forNodes(n,POP) if(n->w!=.0) H-=n->w*::log(n->w);
    return H/MT_LN2;
  }

  /*! entropy $H_i = - \sum_x p(X_i=x) \ln p(X_i=x)$, where $x$ ranges
      from #0# to #maxSymbol#, of the #i#th variable */
  double entropy(const intAL& POP,uint i,int maxSymbol){
    doubleA p;
    marginal(POP,p,i,maxSymbol);
    uint x;
    double H=0.;
    for(x=0;x<p.N;x++) if(p(x)!=.0) H-=p(x)*::log(p(x));
    return H/MT_LN2;
  }

  //! 1D array: entropies for all loci
  void entropy(const intAL& POP,doubleA& H,uint maxLength,int maxSymbol){
    uint i;
    H.resize(maxLength); H=0.;
    for(i=0;i<maxLength;i++) H(i)=entropy(POP,i,maxSymbol);
  }

  //! sum of marginal entropies
  double sumEntropy(const intAL& POP,uint maxLength,int maxSymbol){
    uint i;
    double H=0.;
    for(i=0;i<maxLength;i++) H+=entropy(POP,i,maxSymbol);
    return H;
  }

  /*! mutual information $I_{ij} = \sum_{x,y} p(X_i=x,X_j=y) \ln \frac{
     p(X_i=x,X_j=y) }{p(X_i=x)\, p(X_j=y)}$ between the #i#th and
     #j#th variable */
  double mutualInformation(const intAL& POP,uint i,uint j,int maxSymbol){
    doubleA p,pi,pj;
    pair(POP,p,i,j,maxSymbol);
    eliminate(pi,p,1);
    eliminate(pj,p,0);
    uint x,y;
    double I=0.,pp;
    for(x=0;x<p.d0;x++) for(y=0;y<p.d1;y++){
      pp=p(x,y);
      if(pp!=.0) I+=pp*::log(pp/(pi(x)*pj(y)));
    }
    return I/MT_LN2;
  }

  /*! mutual information $I_{ijk}$ between 3 variables */
  double mutualInformation3(const intAL& POP,uint i,uint j,uint k,int maxSymbol){
    doubleA p,pi,pj,pk;
    triple(POP,p,i,j,k,maxSymbol);
    eliminate(pi,p,1,2);
    eliminate(pj,p,0,2);
    eliminate(pk,p,0,1);
    uint x,y,z;
    double I=0.,pp;
    for(x=0;x<p.d0;x++) for(y=0;y<p.d1;y++) for(z=0;z<p.d2;z++){
      pp=p(x,y,z);
      if(pp!=.0) I+=pp*::log(pp/(pi(x)*pj(y)*pk(z)));
    }
    return I/MT_LN2;
  }

  double mutualInformation(const intAL& POP,uint i,uint j,int maxSymbol,uint conditioned){
    doubleA p,pi,pj,pc;
    marginal(POP,pc,conditioned,maxSymbol);
    uint a,b,c;
    double I=0.,pp;
    for(c=0;c<=(uint)maxSymbol;c++){
      pair(POP,p,i,j,maxSymbol,conditioned,c);
      eliminate(pi,p,1);
      eliminate(pj,p,0);
      for(a=0;a<p.d0;a++) for(b=0;b<p.d1;b++){
	pp=p(a,b);
	if(pp!=.0) I+=pc(c)*pp*::log(pp/(pi(a)*pj(b)));
      }
    }
    return I/MT_LN2;
  }

  //! 2D array: all pair-wise mutual informations (symmetric)
  void mutualInformation(const intAL& POP,doubleA& I,uint maxLength,int maxSymbol){
    uint i,j;
    I.resize(maxLength,maxLength); I=0.;
    for(i=0;i<maxLength;i++) for(j=0;j<i;j++) I(i,j)=I(j,i)=mutualInformation(POP,i,j,maxSymbol);
  }

  //! sum of all pair-wise mutual informations
  double sumMutualInformation(const intAL& POP,uint maxLength,int maxSymbol){
    uint i,j;
    double I=0.;
    for(i=0;i<maxLength;i++) for(j=0;j<i;j++) I+=mutualInformation(POP,i,j,maxSymbol);
    return I;
  }

  //! histogram of the weightings
  void dumpHistogram(const intAL& POP,uint bins=100){
    uintA p;
    p.resize(bins+1); p.setZero();
    intAL::node n;
    uint i;
    forNodes(n,POP){
      i=(uint)floor(n->w*bins);
      if(i>bins) i=bins;
      p(i)++;
    }
    std::ofstream fout("zHisto");
    for(i=0;i<p.N;i++){
      fout <<(.5+i)/bins <<' ' <<p(i) <<'\n';
    }
  }
}

//===========================================================================
//
// analyzing samples
//

struct Loc{ int x; void* p; Loc(){x=-1;p=0;} Loc(int xx){x=xx;p=0;}};
inline std::ostream& operator<<(std::ostream& os,const Loc& l){ os <<l.x; return os; }

//! Pair-wise Compression Coding \ingroup search
class PairCoding{
public:
  MT::List<Loc> P; //the full sequence (concatenation of all samples)
  typedef MT::List<Loc>::node Locus;
  struct Pair{ int c,l,r,n; MT::List_<Locus> L; Pair(){ c=-1; }};
  MT::List<Pair> pairs; //the list of all pairs that occur in the current sequence (and where they occur)

  MT::Array<MT::List<Pair>::node> pairindex; //the `hashtable' mapping a pair to the entry in the pair list

  struct Production{ int c,l,r,n; Production(){ c=-1; n=0; } };

  MT::List<Production> codes;

  uint maxSym;

  struct PairSort{
    bool operator()(MT::List<Pair>::node p,MT::List<Pair>::node q){ return p->L.N > q->L.N; }
  };

  void model(const intAL& POP){ HALT("can't model without encoding"); }
  void encode(const intAL& POP,double abort=0.,double mutateRate=0.){
    int S=0;
    intAL::node n;
    Locus l,ls;//,r;
    MT::List<Pair>::node pn,pm;
    uint i,j;
    Production p;
    MT::Permutation pi;

    P.clear();
    pairs.clear();

    //transform them to lists;
    forNodes(n,POP){
      for(j=0;j<n->N;j++){
	P.append(Loc(n->elem(j)));
	if(n->elem(j)>S) S=n->elem(j);
      }
      P.append(-1);
    }
    maxSym=S;
    //std::cout <<POP;
    //i=0;forNodes(l,P) if(l->x==-1) i++; CHECK(i==POP.N,"dammit");

    //analyze pairs - store them in a list
    pairindex.resize(1000,1000); pairindex.memMove=true; pairindex.setZero();
    //codes.resize(1000);
    forNodes(l,P) newPair(l);

    //double t0=0.,t1=0.,t2=0.,t3=0.,t4=0.;
    for(;;){
      //std::cout <<"iteration " <<k <<" - samples length " <<P.N-POP.N
	//<<" #pairs " <<pairs.N <<std::endl;

      if(abort){
	if(abort<1. && rnd.uni()<abort) break;
        if(abort>=1. && i>=abort) break;
      }

      checkValid();

      MT::getTimer();
      if(!pairs.N) break;
      if((uint)S+1>=pairindex.d0){
	//std::cout <<"symbol limit reached!\n";
	break;
      }

#if 1 //oldest method
      //find max pair
      pi.random(pairs.N);
      for(j=0,i=1;i<pairs.N;i++) if(pairs.nodes(pi(i))->L.N>pairs.nodes(pi(j))->L.N) j=i;
      pm=pairs.nodes(pi(j));
      if(pm->L.N<=1) break;
      replaceAll(pm,S);
#endif

#if 0 //replace pairs -- highest freq first
      pairs.randomIndexing();
      pairs.sortIndexingOp(PairSort());
      if(pairs.nodes(0)->L.N<=1) break;
      //std::cout <<pairs.N <<" pairs:\n";
      j=pairs.N;
      for(i=0;i<j;i++) if(pairs.nodes(i)->L.N>1){
	pm=pairs.nodes(i);
	if(pm->L.N<=1) break;
	if((uint)S+1>=codes.N) break;
	replaceAll(pm,S);
      }

#endif

#if 0 //replace pairs -- in order as they appear in the data

      for(i=0,j=0;i<pairs.N;i++) if(pairs.nodes(i)->L.N>1){ j++; pairs.nodes(i)->c=-2; }else pairs.nodes(i)->c=-1;
      if(!j) break;
      j=pairs.N;
      forNodes(r,P) if(r->p){
	pm=pairindex(r->x,r->next->x);
	if(pm->index<j){
	  if(S+1>=(int)codes.N) break;
	  //does this pair already have a code?
	  if(pm->c==-2){
	    S++;
	    p.c=pm->c=S; p.l=pm->l; p.r=pm->r;
	    codes(p.c)=p;
	  }
	  if(pm->c>0) replaceLocus(r,pm->c); //(pm,S);
	}
      }
#endif

      //t1=MT::getTimer();

      //i=0;forNodes(l,P) if(l->x==-1) i++; CHECK(i==POP.N,"dammit");
      //mutations
      uint m=rnd.poisson(mutateRate * POP.N);
      for(;m--;){
	uint method=rnd(3); //only flips!!! rnd(3)
	switch(method){
	case 0:
	  l=P.rndNode(); if(l->x==-1) break;
	  delPair(l->prev); delPair(l);
  	  l->x=rnd(1+S);
	  newPair(l->prev); newPair(l);
	  break;
	case 1:
	  l=P.rndNode();
	  delPair(l->prev);
	  P.new_node(l,INSbefore);
	  l=l->prev;
  	  l->x=rnd(1+S);
	  newPair(l->prev); newPair(l);
	  break;
	case 2:
	  l=P.rndNode(); if(l->x==-1) break;
	  delPair(l->prev); delPair(l);
	  ls=l->prev;
	  P.del_node(l);
	  l=ls;
	  newPair(l);
	  break;
	}
      }
      //i=0;forNodes(l,P) if(l->x==-1) i++; CHECK(i==POP.N,"dammit");

      //t2=MT::getTimer();

      //delete zero-frequency pairs
      forNodes_save(pn,pm,pairs) if(pn->L.N==0){
	//CHECK(pairindex(pn->l,pn->r)==pn,"dam");
	pairindex(pn->l,pn->r)=0;
	pairs.del_node(pn);
      }

      //t3=MT::getTimer();

      //std::cout <<"times: " <<t0 <<' ' <<t1 <<' ' <<t2 <<' ' <<t3 <<std::endl;
    }
  }

private:
  void checkValid(){
#ifndef MT_NOCHECK
    Locus l;
    MT::List_<Locus>::node ll;
    MT::List<Pair>::node pn;

    //check pairs
    forNodes(pn,pairs){
      forNodes(ll,pn->L){
	l=ll->x;
	CHECK(P.checkNode(l),"not valid");
	CHECK(l->x==pn->l && l->next->x==pn->r,"not valid");
	CHECK(l->p==ll,"not valid");
      }
    }

    //check sequence && pairindex
    forNodes(l,P){
      if(l->x!=-1 && l->next && l->next->x!=-1){
	pn=pairindex(l->x,l->next->x);
	CHECK(pairs.checkNode(pn) && l->x==pn->l && l->next->x==pn->r,"not valid");
	CHECK(pn->L.checkNode((MT::List_<Locus>::node)l->p),"not valid");
      }else{
	CHECK(l->p==0,"not valid");
      }
    }
#endif
  }
  void replaceLocus(Locus l,int code){
    //three pairs are removed: previous, this, next
    delPair(l->prev);
    delPair(l->next);
    delPair(l);
    
    //pair in sequence is replaced
    P.del_node(l->next);
    l->x=code;
    
    //two new pairs
    newPair(l->prev);
    newPair(l);
  }
  void replaceAll(MT::List<Pair>::node pm,int& S){
    Locus l;
    MT::List_<Locus>::node ll,lls;
    Production p;
    
    //replace symbol in sequences
    S++;
    p.c=S; p.l=pm->l; p.r=pm->r; p.n=pm->L.N;
    //std::cout <<"replacements=" <<pm->L.N <<' ';
    forNodes_save(ll,lls,pm->L){ //this is the list of where the pair occurs
      l=ll->x;
      //three pairs are removed: previous, this, next
      delPair(l->prev);
      delPair(l->next);
      lls=ll->next;
      delPair(l);
      
      //pair in sequence is replaced
      P.del_node(l->next);
      l->x=p.c;
      
      //two new pairs
      newPair(l->prev);
      newPair(l);
    }
    codes.append(p);//(p.c)=p;
  }
  void delPair(Locus l){
    MT::List<Pair>::node p;
    if(l && l->x!=-1 && l->next && l->next->x!=-1){
      p=pairindex(l->x,l->next->x);
      p->L.del_node((MT::List_<Locus>::node)l->p);
      l->p=0;
    }
  }
  void newPair(Locus l){
    MT::List<Pair>::node p;
    if(l && l->x!=-1 && l->next && l->next->x!=-1){
      p=pairindex(l->x,l->next->x);
      if(!p){
	p=pairs.new_node();
	pairindex(l->x,l->next->x)=p;
	p->l=l->x; p->r=l->next->x;
      }
      CHECK(p->l==l->x && p->r==l->next->x,"dammit");
      p->L.append(l);
      l->p=p->L.last;
    }
  }


public:
  void decode(intAL& POP){
    Locus l,ls;
    MT::List<Production>::node p;
    forNodes_save(l,ls,P) if(l->x!=-1 && l->x>(int)maxSym){
      p=codes(l->x-maxSym-1);
      CHECK(p->c==l->x,"");
      P.new_node(l,INSafter);
      l->next->x=p->r;
      l->x=p->l;
      ls=l;
    }
    //std::cout <<P;

    uint i=0;
    for(i=0;i<POP.N;i++) POP(i)->clear();
    i=0;
    forNodes(l,P){
      if(l->x!=-1) POP(i)->append(l->x);
      else i++;
    }
  }

  void write(std::ostream& os) const{
    MT::List<Production>::node p;
    os <<"<PairCoding>\n";
    forNodes(p,codes) os <<"  " <<p->c <<':' <<p->l <<' ' <<p->r <<"  ["<<p->n<<"]\n";
    os <<"</PairCoding>" <<std::endl;
  }
};
stdOutPipe(PairCoding);

/*! probabilistic model of samples -- based on only the symbol
    distribution (over all loci) \ingroup search*/
class SymbolDist{
public:
  doubleA lenDist;
  doubleA symDist;

  void model(const intAL& POP){
    uint maxLen,maxSym;

    maxLen=MT::maxLen(POP);
    maxSym=MT::maxSym(POP);

    MT::lengthDist(POP,lenDist,maxLen);

    symDist.resize(maxSym+1); symDist.setZero();
    intAL::node n;
    uint i;
    double Z=0.;
    forNodes(n,POP) for(i=0;i<n->N;i++){ symDist(n->elem(i))+=n->w; Z+=n->w; }
    symDist/=Z;
  }

  void addEntropy(double alpha=-1.,bool alsoLengths=false){
    if(alpha==-1.) alpha=1./lenDist.N;
    if(alsoLengths){
      lenDist.append(0);
      lenDist*=1.-alpha;
      lenDist+=alpha/lenDist.N;
    }

    symDist*=1.-alpha;
    symDist+=alpha/symDist.N;
  }

  void generate(intAL& POP,uint k){
    uint i;
    intAL::node n;
    POP.clear();
    
    //SUS of lengths
    MT::Array<uint> lengths,symbols;
    MT::SUS(lenDist,k,lengths);
    lengths.randomPermute();
    for(i=0;i<k;i++) POP.new_node()->resize(lengths(i));

    forNodes(n,POP) for(i=0;i<n->N;i++) n->elem(i)=MT::SUS(symDist);
  }

  void generateMax(intA& x){
    uint i,l=lenDist.maxIndex();
    x.resize(l);
    for(i=0;i<l;i++) x(i)=symDist.maxIndex();
  }

  double logLikelihood(const intA& x){
    double l=0.;
    uint i;
    for(i=0;i<x.N;i++) if(x(i)!=-1) l+=::log(symDist(x(i)));
    return l/MT_LN2;
  }

  double logLikelihood(const intAL& POP){
    intAL::node n;
    double l=0.;
    forNodes(n,POP) l+=n->w * logLikelihood(n->data());
    return l;
  }

  void write(std::ostream& os) const{
    os <<"length distribution: "<<lenDist
       <<"\nsymbol distribution: "<<symDist;
  }
};
stdOutPipe(SymbolDist);


/*! Factorial probabilistic model (different, independent distributions
  for each locus) \ingroup search*/
class Marginals{
public:
  doubleA lenDist;
  MT::Array<doubleA> symDist;

  void model(const intAL& POP){
    uint maxLen,maxSym;

    maxLen=MT::maxLen(POP);
    maxSym=MT::maxSym(POP);

    MT::lengthDist(POP,lenDist,maxLen);
    MT::marginals(POP,symDist,maxLen,maxSym);
  }

  void addEntropy(double alpha=-1.,bool alsoLengths=false){
    if(alpha==-1.) alpha=1./symDist.N;
    if(alsoLengths){
      lenDist.append(0);
      symDist.resizeCopy(symDist.N+1);
      symDist(symDist.N-1).resize(symDist(0).N);
      symDist(symDist.N-1)=1./symDist(0).N;

      lenDist*=1.-alpha;
      lenDist+=alpha/lenDist.N;
      //lenDist/=sum(lenDist);
    }

    /*
    symDist*=(1.-alpha);
    symDist+=alpha/symDist.d1;
    */
    uint i;
    for(i=0;i<symDist.N;i++){
      symDist(i)*=1.-alpha;
      symDist(i)+=alpha/symDist(i).N;
    }
  }

  void generate(intAL& POP,uint k){
    uint i,j,maxLen=lenDist.N-1;
    intAL::node n;
    POP.clear();
    
    //SUS of lengths
    MT::Array<uint> lengths,symbols;
    MT::SUS(lenDist,k,lengths);
    lengths.randomPermute();
    for(i=0;i<k;i++) POP.new_node()->resize(lengths(i));
    //count how many symbols in the i-th there are
    MT::Array<uint> symCount(maxLen); symCount.setZero();
    forNodes(n,POP) for(i=0;i<n->N;i++) symCount(i)++;
    
    for(i=0;i<maxLen;i++){
      MT::SUS(symDist(i),symCount(i),symbols);
      symbols.randomPermute();
      j=0;
      forNodes(n,POP) if(n->N>i){
	n->elem(i)=symbols(j);
	j++;
      }
    }
  }

  void generateMax(intA& x){
    uint i,l=lenDist.maxIndex();
    x.resize(l);
    for(i=0;i<l;i++) x(i)=symDist(i).maxIndex();
  }

  double logLikelihood(const intA& x){
    CHECK(x.N==symDist.N,"wrong number of marginals");
    double l=0.;
    uint i,L=symDist.N;
    for(i=0;i<L;i++){
      if(x(i)!=-1) l+=::log(symDist(i)(x(i)));
    }
    return l/MT_LN2;
  }

  double logLikelihood(const intAL& POP){
    intAL::node n;
    double l=0.;
    forNodes(n,POP) l+=n->w * logLikelihood(n->data());
    return l;
  }

  void write(std::ostream& os) const{
    os <<"length distribution: "<<lenDist
       <<"\nsymbol distributions: "<<symDist;
  }
};
stdOutPipe(Marginals);




struct TreeNode{ uint i; doubleA p,cp; };
struct TreeEdge{ double I; };
inline std::ostream& operator<<(std::ostream& os,const TreeNode& n){ os<<"var="<<n.i<<" CPT="<<n.p; return os; }
inline std::ostream& operator<<(std::ostream& os,const TreeEdge& e){ os<<e.I; return os; }

/*! Tree probabilistic model \ingroup search*/
class Tree:public Graph<TreeNode,TreeEdge>{
public:
  struct CompareOp{ bool operator()(Graph<TreeNode,TreeEdge>::edge a,Graph<TreeNode,TreeEdge>::edge b) const{ return a->I>b->I; } };
  void model(const intAL& POP){
    node n;
    edge e;
    uint L=MT::maxLen(POP),S=MT::maxSym(POP),i,j;
    clear();
    MT::Permutation pi; pi.random(L);
    for(i=0;i<L;i++){
      n=new_node();
      n->i=pi(i);
      //p.marginal(n->marg,n->i,S);
      //n->H=entropy(n->marg);
    }
    for(i=0;i<L;i++) for(j=0;j<i;j++){
      e=new_edge(nodes(i),nodes(j));
      e->I=MT::mutualInformation(POP,e->from->i,e->to->i,S);
    }

    maximumSpanningTree(*this,CompareOp());
    topsort();
    for_nodes(n){
      n->index=n->i; nodes(n->index)=n;
      //for_out_edges_save(e,es,n) if(e->I < 1e-1) del_edge(e);
    }


    for_nodes(n){
      if(n->Nin==0){
	MT::marginal(POP,n->p,n->i,S);
      }
      if(n->Nin==1){
        MT::pair(POP,n->p,n->firstIn->from->i,n->i,S);
        for(i=0;i<=S;i++) normalizeDist(n->p[i]());
      }
    }
  }

  void addEntropy(double alpha=-1.){
    if(alpha==-1.) alpha=1./N;
    node n;
    for_nodes(n){
      if(n->Nin==0){
	n->p*=1.-alpha;
	n->p+=alpha/n->p.N;
      }
      if(n->Nin==1){
	n->p*=1.-alpha;
	n->p+=alpha/n->p.d1;
      }
    }
  }

  void checkValid(){
    node n;
    uint i;
    double Z;
    for_nodes(n){
      if(n->Nin==0){
        Z=sum(n->p);
        CHECK(fabs(Z-1.)<1e-6,"tree check failed");
      }
      if(n->Nin==1){
	for(i=0;i<n->p.d0;i++){
	  Z=sum(n->p[i]);
	  CHECK(fabs(Z-1.)<1e-6,"tree check failed");
	}
      }
    }
  }

  void generate(intAL& POP,uint k){
    POP.clear();
    intAL::node x;
    node n;
    uint i,j;
    for(i=0;i<k;i++){
      x=POP.new_node();
      x->w=1./k;
      x->resize(N);
      for_nodes(n){
	if(n->Nin==0){
          x->elem(n->i)=MT::SUS(n->p);
	}
	if(n->Nin==1){
	  j=n->firstIn->from->i;
	  x->elem(n->i)=MT::SUS(n->p[x->elem(j)]);
	}
      }
    }
  }

  void generateMax(intA& x){
    node n;
    uint j;
    x.resize(N);
    for_nodes(n){
      if(n->Nin==0){
        x(n->i)=n->p.maxIndex();
      }
      if(n->Nin==1){
        j=n->firstIn->from->i;
        x(n->i)=n->p[x(j)].maxIndex();
      }
    }
  }

  void random(uint N,double eps=.75){
    clear();
    node n;
    edge e;
    uint i;
    for(i=0;i<N;i++){
      n=new_node();
      n->i=i;
      if(!i){
	n->p.resize(2); n->p=.5;
      }else{
	e=new_edge(nodes(rnd(i)),n);
        n->p.resize(2,2);
	n->p(0,0)=eps;    n->p(0,1)=1.-eps;
	n->p(1,0)=1.-eps; n->p(1,1)=eps;
      }
    }
  }

  double logLikelihood(const intA& x){
    CHECK(x.N==N,"wrong number of nodes");
    node n,m;
    double l=0.,p;
    uint j,i;

#if 1
    for_nodes(n){
      n->cp.resize(first->p.N); n->cp.setZero();
      if(x(n->i)!=-1){
	n->cp(x(n->i))=1.;
      }else{
	if(n->Nin==0){
	  n->cp=n->p;
	}
	if(n->Nin==1){
	  m=n->firstIn->from;
	  for(i=0;i<n->p.d0;i++) n->cp+=!(m->cp(i)*n->p[i]);
	}
	CHECK(fabs(sum(n->cp)-1.)<1e-6,"");
	//CHECK(n->cp(0)>0. && n->cp(1)>0.,"");
      }
    }
    //std::cout <<x <<'\n' <<!~mp;
#endif
    
    for_nodes(n) if(x(n->i)!=-1){
      if(n->Nin==0){
	l+=::log(n->p(x(n->i)));
      }
      if(n->Nin==1){
	m=n->firstIn->from;
	j=n->firstIn->from->i;
	if(x(j)!=-1){
	  l+=::log(n->p(x(j),x(n->i)));
	}else{
  	  for(p=0.,i=0;i<n->p.d0;i++) p+=m->cp(i)*n->p(i,x(n->i));
	  CHECK(p>0,"");
	  l+=::log(p);
	}
      }
    }
    return l/MT_LN2;
  }
  
  double logLikelihood(const intAL& POP){
    intAL::node n;
    double l=0.;
    forNodes(n,POP) l+=n->w * logLikelihood(n->data());
    return l;
  }
};


/*! 3rd dependencies graphical model \ingroup search*/
class TriTree:public Graph<TreeNode,TreeEdge>{
public:
  doubleA I2,I3;

  struct CompareOp{ bool operator()(Graph<TreeNode,TreeEdge>::edge a,Graph<TreeNode,TreeEdge>::edge b) const{ return a->I>b->I; } };
  void model(const intAL& POP){
    node n,nn,mn,mnn;
    edge e;
    uint L=MT::maxLen(POP),S=MT::maxSym(POP),i,j,k; int m;
    boolA done(L); done=false;
    double mI,I;

    //calculate triple-wise mutual information
    I2.resize(L,L); I2=0.;
    for(i=0;i<L;i++) for(j=0;j<i;j++){
      I2(i,j)=MT::mutualInformation(POP,i,j,S);
    }

    I3.resize(L,L,L); I3=0.;
    for(i=0;i<L;i++) for(j=0;j<i;j++) for(k=0;k<j;k++){
      I3(i,j,k)=MT::mutualInformation3(POP,i,j,k,S);
    }

    //construct graph (similar to maximum spanning tree)
    clear();
    MT::Permutation pi; pi.random(L);

    //root
    n=new_node();
    n->i=rnd(L);
    done(n->i)=true;

    //second, based on I2
    m=-1;
    mn=0;
    mI=0.;
    for(i=0;i<L;i++){
      j=pi(i);
      if(!done(j)){
	for_nodes(n){
	  if(m==-1 || MI2(n->i,j)>mI){ mI=MI2(n->i,j); m=j; mn=n; }
	}
      }
    }
    n=new_node();
    n->i=m;
    done(n->i)=true;
    e=new_edge(mn,n); e->I=0;

    //rest, based on I3
    while(N<L){
      m=-1;
      mn=mnn=0;
      mI=0.;
      for(i=0;i<L;i++){
	j=pi(i);
	if(!done(j)){
	  for_nodes(n) for_nodes(nn) if(n!=nn){
	    I=MI3(n->i,nn->i,j)-MI2(n->i,nn->i);
	    CHECK(I>-1e-6,"dammit");
	    if(m==-1 || I>mI){ mI=I; m=j; mn=n; mnn=nn; }
	  }
	}
      }
      n=new_node();
      n->i=m;
      done(n->i)=true;
      e=new_edge(mn,n); e->I=0;
      e=new_edge(mnn,n); e->I=0;
    }

    //sort topology and index according to variables
    bool ts=topsort();
    CHECK(ts,"dammit");
    for_nodes(n){ n->index=n->i; nodes(n->index)=n; }

    //calculate conditional probability tables
    for_nodes(n){
      if(n->Nin==0){
        MT::marginal(POP,n->p,n->i,S);
      }
      if(n->Nin==1){
        j=n->firstIn->from->i;
        MT::pair(POP,n->p,j,n->i,S);
        for(i=0;i<=S;i++) normalizeDist(n->p[i]());
      }
      if(n->Nin==2){
	j=n->firstIn->from->i;
	k=n->lastIn->from->i;
        MT::triple(POP,n->p,j,k,n->i,S);
        for(i=0;i<=S;i++) for(j=0;j<=S;j++) normalizeDist(n->p[i][j]());
      }
    }
  }

  double MI2(uint i,uint j){
    uint l;
    if(i<j){ l=i; i=j; j=l; }
    return I2(i,j);
  }
  double MI3(uint i,uint j,uint k){
    uint l;
    if(i<j){ l=i; i=j; j=l; }
    if(i<k){ l=i; i=k; k=l; }
    if(j<k){ l=j; j=k; k=l; }
    return I3(i,j,k);
  }

  void addEntropy(double alpha=-1.){
    if(alpha==-1.) alpha=1./N;
    node n;
    for_nodes(n){
      if(n->Nin==0){
	n->p*=1.-alpha;
	n->p+=alpha/n->p.N;
      }
      if(n->Nin==1){
	n->p*=1.-alpha;
	n->p+=alpha/n->p.d1;
      }
      if(n->Nin==2){
	n->p*=1.-alpha;
	n->p+=alpha/n->p.d2;
      }
    }
  }

  void generate(intAL& POP,uint Pn){
    POP.clear();
    intAL::node x;
    node n;
    uint i,j,k;
    for(i=0;i<Pn;i++){
      x=POP.new_node();
      x->w=1./Pn;
      x->resize(N);
      for_nodes(n){
	if(n->Nin==0){
          x->elem(n->i)=MT::SUS(n->p);
	}
	if(n->Nin==1){
	  j=n->firstIn->from->i;
	  x->elem(n->i)=MT::SUS(n->p[x->elem(j)]);
	}
	if(n->Nin==2){
	  j=n->firstIn->from->i;
	  k=n->lastIn->from->i;
	  x->elem(n->i)=MT::SUS(n->p[x->elem(j)][x->elem(k)]);
	}
      }
    }
  }

  void generateMax(intA& x){
    node n;
    uint j,k;
    x.resize(N);
    for_nodes(n){
      if(n->Nin==0){
        x(n->i)=n->p.maxIndex();
      }
      if(n->Nin==1){
        j=n->firstIn->from->i;
        x(n->i)=n->p[x(j)].maxIndex();
      }
      if(n->Nin==2){
        j=n->firstIn->from->i;
        k=n->lastIn->from->i;
        x(n->i)=n->p[x(j)][x(k)].maxIndex();
      }
    }
  }

  void random(uint N){
    clear();
    node n;
    edge e;
    uint i;
    for(i=0;i<N;i++){
      n=new_node();
      n->i=i;
      if(!i){
	n->p.resize(2); n->p=.5;
      }else{
	e=new_edge(nodes(rnd(i)),n);
        n->p.resize(2,2);
	n->p(0,0)=.75; n->p(0,1)=.25;
	n->p(1,0)=.25; n->p(1,1)=.75;
      }
    }
  }

  double logLikelihood(const intA& x){
    CHECK(x.N==N,"wrong number of nodes");
    node n;
    double l=0.;
    uint j,k;

    for_nodes(n) if(x(n->i)!=-1){
      if(n->Nin==0){
	l+=::log(n->p(x(n->i)));
      }
      if(n->Nin==1){
	j=n->firstIn->from->i;
	l+=::log(n->p(x(j),x(n->i)));
      }
      if(n->Nin==2){
	j=n->firstIn->from->i;
	k=n->lastIn->from->i;
	l+=::log(n->p(x(j),x(k),x(n->i)));
      }
    }
    return l/MT_LN2;
  }

  double logLikelihood(const intAL& POP){
    intAL::node n;
    double l=0.;
    forNodes(n,POP) l+=n->w * logLikelihood(n->data());
    return l;
  }
};


/*! Walsh analysis \ingroup search*/
class Walsh{
public:
  struct Coeff{ doubleA w; TupleIndex loci; };
  MT::Array<Coeff> W;

  double coeff(uint i){
    uint n=0;
    while(i){ if(i&1) n++; i=i>>1; }
    if(n&1) return -1.; else return 1.;
  }

  void model(const intAL& POP,uint order=4){
    uint a,i,k,L=MT::maxLen(POP),S=MT::maxSym(POP);
    CHECK(S==1,"Walsh only for binary samples");
    doubleA p;

    //doubleA I; POP.mutualInformation(I,L,S); std::cout <<"mutual Info=" <<I;

    if(order>L) order=L;

    double w,x,y;
    W.resize(order+1);
    for(k=0;k<=order;k++){
      std::cout <<"Walsh order "<<k<<" ..." <<std::endl;
      W(k).loci.init(k,L);
      W(k).w.resize(W(k).loci.d0);
      if(!k) W(0).w(0)=1.; else
      for(a=0;a<W(k).w.N;a++){
	MT::tuple(POP,p,W(k).loci[a],S);
	w=0.; x=y=1.;
	for(i=0;i<p.N;i++){
	  w+=coeff(i)*p(i);
	  if(coeff(i)>0.) x*=p(i); else y*=p(i);
	}
	//if(x) w=2./(1.+(y/x))-1.; else if(y) w=-1.; else w=0.;
	W(k).w(a)=w;
      }
    }
  }

  void spectrum(doubleA& avg,doubleA& max) const{
    avg.resize(W.N); avg.setZero();
    max.resize(W.N); avg.setZero();
    uint k,a;
    for(k=0;k<W.N;k++){
      for(a=0;a<W(k).w.N;a++) avg(k)+=fabs(W(k).w(a));
      avg(k)/=W(k).w.N;
      max(k)=W(k).w(W(k).w.maxIndex());
    }
  }

  void dumpSpec(char* name){
    uint a,k;
    std::ofstream fout;
    MT::open(fout,name);
    for(k=0;k<W.N;k++) for(a=0;a<W(k).w.N;a++){
      fout <<-.25 + k + .5*a/W(k).w.N  <<' ' <<W(k).w(a) <<' ' <<fabs(W(k).w(a)) <<std::endl;
    }
  }

  void write(std::ostream& os) const{
#if 0
    uint a;
    uint k;
    for(k=0;k<W.N;k++){
      os <<k <<"th order w's =\n";
      for(a=0;a<W(k).w.N;a++){ if(k) W(k).loci[a].write(os,true); os <<" : " <<W(k).w(a) <<std::endl; }
    }
#endif
    doubleA s,m; spectrum(s,m);
    os <<"spectrum = " <<s <<m <<std::endl;
  }
};
stdOutPipe(Walsh);



struct Code{
  int v;
  uintA loci;
};
inline std::ostream& operator<<(std::ostream& os,const Code& c){ return os <<c.v <<c.loci; }
//! Split Coding \ingroup search
class SplitCoding{
public:
  uint length;
  MT::Array<Code> codes;
  doubleA marginals;

  void checkValid(){
    uint ml=marginals.d0;
    Code* c;
    forAll(c,codes)
      CHECK(c->loci(0)<ml &&
	    c->loci(1)<ml &&
	    c->loci(2)<ml,"code invalid");
  }

  void model(intAL& POP,uint k=1){
    Code c;

    uint i,j,L=MT::maxLen(POP),S=MT::maxSym(POP);
    length=L;
    //doubleA I;
    //p.mutualInformation(I,L,S);
    //std::cout <<std::setprecision(3) <<"mutual information after iteration " <<0 <<": \n" <<I <<sum(I);

    codes.clear();
    for(;k--;){
      Tree t;
      Tree::node v;
      t.model(POP);
      //std::cout <<t;
      //Tree::edge e,es;
      //forNodes(v,t) for_out_edges_save(e,es,v) if(e->I<.01) t.del_edge(e);
      forNodes_rev(v,t){
	if(v->Nin==1){
	  i=v->firstIn->from->i;
	  j=v->i;
	  
	  c.v=1;
	  c.loci.resize(3);
	  c.loci(0)=i; c.loci(1)=j; c.loci(2)=L;

	  /* //oder implementation
	  forNodes(n,POP){
	    CHECK(n->N==L,"dammit");
	    n->append(-1);
	  }
	  inverseapply(c,p);
	  */

	  L++;
	  codes.append(c);
	}
      }
    }
    MT::marginals(POP,marginals,L,S);
  }

  void encode(intA& x,bool sample=false){
    CHECK(x.N==length || x.N==length+codes.N,"have you modelled the distribution before?");
    uint i;
    if(x.N==length){
      x.resizeCopy(length+codes.N);
      for(i=length;i<x.N;i++) x(i)=-1;
    }
    for(i=0;i<codes.N;i++){
      CHECK(codes(i).loci(0)<x.N &&
	    codes(i).loci(1)<x.N && 
	    codes(i).loci(2)<x.N,"invalid code");
      if(x(codes(i).loci(0))==codes(i).v) x.permute(codes(i).loci(1),codes(i).loci(2));
    }
    if(sample){
      for(i=0;i<x.N;i++) if(x(i)==-1) x(i)=MT::SUS(marginals[i]);
    }
  }

  void encode(intAL& POP,bool sample=true){
    intAL::node n;
    forNodes(n,POP) encode(n->data(),sample);
  }

  /* //older implementation
  void inverseapply(Code& c,intAL& POP){
    //std::cout <<"expanding variable "<<c.loci(1)<<" to "<<c.loci(1)<<"-"<<c.loci(2)<<" conditioned on "<<c.loci(0)<<std::endl;
    intAL::node n;
    forNodes(n,POP){
      if(n->elem(c.loci(0))==c.v) n->permute(c.loci(1),c.loci(2));
    }
  }*/

  void decode(intA& x){
    uint i;
    for(i=codes.N;i--;){
      CHECK(codes(i).loci(0)<x.N &&
	    codes(i).loci(1)<x.N && 
	    codes(i).loci(2)<x.N,"invalid code");
      if(x(codes(i).loci(0))==codes(i).v) x.permute(codes(i).loci(1),codes(i).loci(2));
    }
    //x.resizeCopy(length);
  }

  void decode(intAL& POP){
    intAL::node n;
    forNodes(n,POP) decode(n->data());
  }

  void generate(intAL& POP,uint k){
    POP.clear();
    intAL::node n;
    uint i,j,L=marginals.d0;
    for(i=0;i<k;i++){
      n=POP.new_node();
      n->w=1./k;
      n->resize(L);
      for(j=0;j<L;j++) n->elem(j)=MT::SUS(marginals[j]);
    }

    decode(POP);
  }

  void generateMax(intA& x){
    uint j,L=marginals.d0;
    x.resize(L);
    for(j=0;j<L;j++) x(j)=marginals[j].maxIndex();

    decode(x);

    x.resizeCopy(length);
  }

  void write(std::ostream& os) const{
    os <<codes.N <<" expansions:\n";
    for(uint i=0;i<codes.N;i++){
      os <<codes(i).loci(1) <<" expanded to "<<codes(i).loci(1)<<"-"<<codes(i).loci(2)<<" conditioned on "<<codes(i).loci(0)<<"\n";
    }
    os <<"marginals over "<<marginals.d0<< " expanded variables:\n" <<marginals <<std::endl;
  }
};
stdOutPipe(SplitCoding);

//! Amari's theta coordinates \ingroup search
class Amari{
public:
  doubleA theta0,theta1,theta2,theta3;
  TupleIndex idx2,idx3;

  void model(const intAL& POP){
    uint a,i,j,k,L=MT::maxLen(POP),S=MT::maxSym(POP);
    CHECK(S==1,"Amari only for binary samples");
    doubleA p;
    double x,y;

    doubleA I; MT::mutualInformation(POP,I,L,S);
    std::cout <<"mutual Info=" <<I;

    //2nd order
    idx2.init(2,L);
    theta2.resize(idx2.d0); theta2=0.;
    for(a=0;a<theta2.N;a++){
      i=idx2(a,0); j=idx2(a,1);
      MT::pair(POP,p,i,j,S);
      x=p(1,1)*p(0,0);
      y=p(1,0)*p(0,1);
      //theta2(i,j)=::log(x/y);             //original
      if(x) theta2(a)=2./(1.+(y/x)) - 1.; //sigmoidal
      else{ if(y) theta2(a)=-1.; else theta2(a)=0.; }
    }

    //3rd order
    idx3.init(3,L);
    theta3.resize(idx3.d0); theta3=0.;
    for(a=0;a<theta3.N;a++){
      i=idx3(a,0); j=idx3(a,1); k=idx3(a,2);
      MT::triple(POP,p,i,j,k,S);
      x=p(1,1,1)*p(1,0,0)*p(0,1,0)*p(0,0,1);
      y=p(1,1,0)*p(1,0,1)*p(0,1,1)*p(0,0,0);
      //theta3(i,j,k)=::log(x/y);             //original
      if(x) theta3(a)=2./(1.+(y/x)) - 1.; //sigmoidal
      else{ if(y) theta3(a)=-1.; else theta3(a)=0.; }
    }
  }

  void write(std::ostream& os) const{
    uint a;
    os <<"2nd order theta's =\n";
    for(a=0;a<theta2.N;a++){ os <<idx2[a].ioraw() <<" : " <<theta2(a) <<std::endl; }
    os <<"3rd order theta's =\n";
    for(a=0;a<theta3.N;a++){ os <<idx3[a].ioraw() <<" : " <<theta3(a) <<std::endl; }
  }
};
stdOutPipe(Amari);


class Gaussian{
public:
  double sdv;
  doubleA mean,cov,eig,U;

  void clear(){ sdv=0; mean.clear(); cov.clear(); eig.clear(); U.clear(); }

  void model(doubleAL& L,bool zeroMean=false){
    doubleAL::node n;
    LinearStatistics S;
    forNodes(n,L) S.learn(n->data(),n->w);
    if(zeroMean) S.computeZeroMean(); else S.compute();
    mean=S.MeanX;
    cov=S.VarX;
    calcDecomp();
  }

  void calcDecomp(){
    doubleA V;
    MT::svd(cov,U,V,eig);
    double var=sum(eig);
    if(var>1e-10){
      for(uint i=0;i<eig.N;i++) eig(i)=sqrt(eig(i)/var*eig.N); //trace of eig^2 becomes be N!
      sdv=sqrt(var);
    }else{
      eig=1.;
      sdv=0.;
    }
  }

  void generate(doubleAL& L,uint k){
    uint i,j;
    doubleAL::node n;
    L.clear();
    for(i=0;i<k;i++){
      n=L.new_node();
      n->resize(mean.N);
      rndGauss(n->data(),sdv);
      for(j=0;j<n->N;j++) (n->data())(j)*=eig(j);
      n->data()=U*(n->data());
      n->data()+=mean;
      n->w=1./k;
    }
  }

  double prob(doubleA& x){
    doubleA y=x;
    y-=mean;
    y=~U*y;
    for(uint j=0;j<y.N;j++) y(j)/=eig(j);
    return ::exp(-.5*sumOfSqr(y)/(sdv*sdv));
  }

  void addEntropy(double eps=1.1){
    sdv*=eps;
    for(uint i=0;i<eig.N;i++) eig(i)+=pow(eig(i),.8);
    eig/=sqrt(sumOfSqr(eig)/eig.N);
  }

  void report(std::ostream& os){
    os <<"mean=" <<mean <<"\nsdv=" <<sdv <<"\neig=" <<eig <<"\nU=" <<U <<'\n';
  }
};







#endif
