#ifndef MT_search_h
#define MT_search_h

#include"distributions.h"
#include"codeMap.h"

namespace MT{

/*! \defgroup search Search Module

  The underlying data structure in all search schemes is the \ref
  MT::intAL . It stores a list of integer \ref MT::intA's and
  allows to associate weights (or probabilities) to them. I think of
  it as a particle representation of a distribution. It provides basic
  calculations for such a distribution, like calculating symbol
  marginals, entropies or mutual informations, and allows selection
  and resampling of the distribution.

  More complex classes build probabilistic models (marginals, trees,
  graphical models, etc) from such a population. Also Walsh analysis
  is implemented. (see the \ref distribution.h file).

  Search heuristics receive a weighted (=evaluated) population as
  input and return a population of (yet unevaluated) new search
  samples. I implemented several serach heuristics, each of them only
  a few lines of code. They are meant as templates for user-defined,
  more tailored search heuristics. See the compounds below.

  In principle, search means to iterate calling a search heuristic and
  evaluating the whole population of samples. The \ref Search class
  is utility class for this iteration. It allows to combine search
  with hill climbing, specify stopping times, writes performance data
  into files, repeates search several times on a problem to get
  statistics, etc. It doesn't do anything smart.
*/
//@{

//===========================================================================
//
// basic search ingredients
//

/*! hill climbs the solution x w.r.t. the given evaluation -- currently only for binary strings! */
template<class Evaluation>
double hillClimber(intA& x,Evaluation& eval){
  uint n=x.N;
  MT::Permutation pi;
  uint i;
  double curV,sampleV;
  curV=eval(x);            //current value
  for(;;){
    pi.random(n);
    for(i=0;i<n;i++){
      x(pi(i))^=1;         //single bit flip alteration
      sampleV=eval(x);     //new value
      if(sampleV==1.) return sampleV; //found the optimum;
      if(sampleV>curV){ curV=sampleV; break; } //better -> keep it
      //if(sampleV==curV) if(rnd(2)) break;    //equally good - random accept
      x(pi(i))^=1;         //undo alteration
    }
    //std::cout <<"climb " <<curV <<" " <<i <<std::endl;
    if(i==n) return curV;  //no improvements
  }
}

template<class Evaluation>
double hillClimber(doubleA& x,Evaluation& eval){
  HALT("is not implemented yet!");
}

/*! almost like a hill climer */
template<class T,class Evaluation>
double gsat(MT::Array<T>& x,Evaluation& eval){
  uint n=x.N;
  uint i;
  double curV,sampleV,bestV;
  MT::List_<uint> loci;
  curV=bestV=eval(x);
  for(;;){
    loci.clear();
    for(i=0;i<n;i++){
      x(i)^=1;		    //single bit flip alteration
      sampleV=eval(x);      //new value
      if(sampleV==1.) return sampleV; //found the optimum
      if(sampleV>bestV){
        loci.clear();
        loci.append(i);
        bestV=sampleV;
      }
      if(sampleV==bestV && sampleV>curV){
	loci.append(i);
      }
      x(i)^=1;		  //undo alteration
    }
    if(!loci.N) return curV; //no improvements
    i=loci.rndNode()->x;
    x(i)^=1;              //do best alteration
    curV=bestV;     //current value
    //std::cout <<"climb " <<curV <<" " <<i <<std::endl;
  }
}

/*! one point mutation of all samples in the population */
inline void mutateOnePoint(intAL& p,int low=0,int hi=1){
  intAL::node n;
  forall(n,p) if(n->N) n->elem(rnd(n->N))=low+rnd(1+hi-low);
};

/*! mutate the string x */
inline uint mutate(intA& x,int low=0,int hi=1,double rate=1.,bool flipOnly=true,bool stepsOnly=false){
  uint m,M=rnd.poisson(rate);
  byte method;
  for(m=0;m<M;m++){
    if(flipOnly) method=0; else method=rnd(3);
    switch(method){
    case 0:
      if(!stepsOnly){
        if(x.N) x.elem(rnd(x.N))=low+rnd(1+hi-low);
      }else{
        if(x.N) x.elem(rnd(x.N))+=rnd(2)?-1:1;
      }
      break;
    case 1:
      x.insert(rnd(x.N+1),low+rnd(1+hi-low));
      break;
    case 2:
      if(x.N) x.remove(rnd(x.N),1);
      break;
    }
  }
  return M;
}

//! mutate whole population
inline void mutate(intAL& p,int low=0,int hi=1,double rate=1.,bool flipOnly=true,bool stepsOnly=false){
  intAL::node n;
  forall(n,p) mutate(n->data(),low,hi,rate,flipOnly,stepsOnly);
}

//! struct wrapper for the mutate function
struct MutationOp{
  static void mutate(intA& x,int low=0,int hi=1,double rate=1.){ MT::mutate(x,low,hi,rate,false); }
};

/*! uniform crossover between two strings */
inline void cross(intA& x,intA& y){
  uint i,N=MT::MIN(x.N,y.N);
  int a;
  for(i=0;i<N;i++) if(rnd(2)){ a=x(i); x(i)=y(i); y(i)=a; }
}

/*! uniform crossover for a whole population */
inline void cross(intAL& p){
  p.randomIndexing();
  uint i;
  for(i=0;i<p.N;i+=2) cross(p.nodes(i)->data(),p.nodes(i+1)->data());
}


//===========================================================================
//
// some search heuristics (e.g. for use with the Search class)
//

//! random search -- produces 'lambda' random sequences of length 'length' at every iteration
struct RandomSearch{
  uint
    lambda, //!< #samples in each generation [1]
    length; //!< length of samples [1]
  int
    lo, //!< lowest symbol [0]
    hi; //!< highest symbol [1]
  RandomSearch(){ lambda=1; length=1; lo=0; hi=1; }
  void operator()(intAL& sample,std::ostream *evoFile=0){
    random(sample,lambda,length,lo,hi);
  }
};

//! standard GA
struct PlainGA{
  uint
    mu,	    //!< parent size [default 30]
    lambda, //!< offspring size [default 100]
    initL;  //!< initialization string length [10]
  intA
    E,	    //!< elitist
    *init;  //!< initialization sequence [default 0 -> random initialization]
  int
    lo,	    //!< minimal symbol [default 0]
    hi;	    //!< maximal symbol [default 1]
  bool
    crossover, //!< use crossover [true]
    varLength; //!< variable length mutations [true]
  //! constructor
  PlainGA(){ mu=30; lambda=100; lo=0; hi=1; init=0; initL=10; crossover=varLength=true; }
  //! heuristic core: (1) select parents (2) resample offspring (3) cross them (4) mutate them (5) readd elitist
  void operator()(intAL& sample,std::ostream *evoFile=0){
    if(!sample.N){ if(init) sample.set(lambda,*init); else random(sample,lambda,initL,lo,hi); return; }

    sample.selectBest(mu);
    E=sample(0)->data();//elitist
    sample.weightEqual();

    sample.resample(lambda);
    if(crossover) cross(sample);
    mutate(sample,lo,hi,1.,!varLength);
    //mutateOnePoint(sample,lo,hi);
    sample.add(E);
  }
};

//! GA with pair-wise compression coding
struct PairCodeGA{
  CodeMap map;
  uint mu,lambda,initL;
  intA E,*init;
  int lo,hi;
  PairCodeGA(){ mu=30; lambda=100; lo=0; hi=1; init=0; initL=1; }
  void operator()(intAL& sample,std::ostream *evoFile=0){
    if(!sample.N){ if(init) sample.set(lambda,*init); else random(sample,lambda,initL,lo,hi); return; }

    sample.selectBest(mu);
    E=sample(0)->data();//elitist
    sample.weightEqual();

    sample.resample(lambda);
    map.encodeAndMute(sample,MutationOp(),lo,hi);
    map.decode(sample);

    sample.add(E);
  }
};

#if 0
//! GA with Split Coding
struct SplitCodeGA{
  SplitCoding map;
  uint mu,lambda,initL;
  intA E,*init;
  SplitCodeGA(){ mu=30; lambda=100; init=0; initL=1; }
  void operator()(intAL& sample,std::ostream *evoFile=0){
    if(!sample.N){ if(init) sample.set(lambda,*init); else random(sample,lambda,initL,0,1); return; }

    sample.selectBest(mu);
    E=sample(0)->data();//elitist
    sample.weightEqual();

    map.model(sample);
    map.encode(sample);
    sample.resample(lambda);
    cross(sample);
    mutate(sample);
    map.decode(sample);
    sample.add(E);
  }
};
#else
//! another Split Code GA
struct SplitCodeGA{
  SplitCoding map;
  intAL pool,selpool;
  uint mu,lambda,poolSize,initL,counter;
  double selRate;
  intA E;	    //!< elitist
  bool crossover;
  SplitCodeGA(){ mu=30; lambda=100; poolSize=1000; selRate=.1; initL=1; counter=0; crossover=false; }
  void operator()(intAL& sample,std::ostream *evoFile=0){
    if(!sample.N){ random(sample,lambda,initL,0,1); pool.clear(); counter=0; return; }

    //collect all previous samples (truncated only)
    pool.insert(sample);
    intAL::node n; forall(n,pool) if(n->N!=initL) n->resizeCopy(initL);
    pool.makeDisjoint();
    if(pool.N>poolSize){
      //pool.selectBest(poolSize);    //keep the best in the pool
      uint d=pool.N-poolSize;	    //or keep the most recent in the pool
      for(;d--;) pool.del_node(pool.first);
    }

    //recalculate a split coding from the best-so-far samples every 20 generations
    if(!counter || !(counter%100)){
      selpool.copyBest(pool,uint(selRate*pool.N)+10);
      selpool.weightEqual();
      
      map.model(selpool);
      intAL::node n; forall(n,sample) n->resizeCopy(initL); //truncate current samples
      //counter=1;
    }
    counter++;

    //select, encode, mutate, decode current samples
    sample.selectBest(mu);
    E=sample(0)->data();
    map.encode(sample);
    sample.resample(lambda);
    if(crossover) cross(sample);
    mutate(sample,0,1,10.,true);
    map.decode(sample);
    sample.add(E);
  }
};
#endif

/*! EDA -- maintains a pool of the last `poolSize' best (disjoint) solutions
    found so far; at each iteration, fits a probabilistic model to this pool,
    adds entropy, and samples from it */
template<class Model>
struct PoolEDA{
  Model model;
  intAL pool;
  uint poolSize,lambda,initL;
  int lo,hi;
  double selRate,entropy;
  PoolEDA(){ poolSize=1000; selRate=.1; entropy=-1.; lambda=100; lo=0; hi=1; initL=1; }
  void operator()(intAL& sample,std::ostream *evoFile=0){
    if(!sample.N){
      random(sample,lambda,initL,lo,hi); pool.clear();
      if(evoFile) (*evoFile) <<"0 0 0 ";
      return;
    }

    //sample.selectBest(mu);	    //add only selected to the pool
    pool.unite(sample);
    pool.makeDisjoint();
    if(pool.N>poolSize){
      //pool.selectBest(poolSize);    //keep the best in the pool
      uint d=pool.N-poolSize;	    //or keep the most recent in the pool
      for(;d--;) pool.del_node(pool.first);
    }
    sample.copyBest(pool,uint(selRate*pool.N)+10); //build model only from best 30%
    //sample.copyList(pool);	    //or build model from all
    sample.weightEqual();

    model.model(sample);
    model.addEntropy(entropy);
    model.generate(sample,lambda);
    CHECK(sample.N==lambda,"sample generation failed");

    if(evoFile) (*evoFile) <<pool.N <<' ' <<pool(0)->w <<' ' <<pool.average() <<' ';
  }
};

/*! EDA -- maintains a pool of the last `poolSize' best (disjoint) solutions
    found so far; at each iteration, fits a probabilistic model to this pool,
    adds entropy, and samples from it */
struct PairCodeEDA{
  Marginals model;
  CodeMap map;
  intAL pool;
  uint poolSize,mu,lambda,initL;
  bool usePool,varLen;
  int lo,hi;
  double selRate,beta;
  PairCodeEDA(){ poolSize=1000; selRate=.1; beta=.1; lambda=100; lo=0; hi=1; initL=2; usePool=true; varLen=true; }
  void operator()(intAL& sample,std::ostream *evoFile=0){
    if(!sample.N){
      random(sample,lambda,initL,lo,hi); pool.clear();
      if(evoFile) (*evoFile) <<"0 0 0 ";
      return;
    }

    if(usePool){
      pool.unite(sample);
      pool.makeDisjoint();
      if(pool.N>poolSize){
	//uint d=pool.N-poolSize;	    //or keep the most recent in the pool
	//for(;d--;) pool.del_node(pool.first);
	pool.selectBest(poolSize);
      }
      sample.copyBest(pool,uint(selRate*pool.N)+10); //build model only from best ?%
      if(evoFile) (*evoFile) <<pool.N <<' ' <<pool(0)->w <<' ' <<pool.average() <<' ';
      sample.weightEqual();
    }else{
      sample.selectBest(uint(selRate*sample.N)); //build model only from best ?%
      if(evoFile) (*evoFile) <<sample.N <<' ' <<sample(0)->w <<' ' <<sample.average() <<' ';
      sample.weightEqual();
    }

    map.encode(sample,beta);
    model.model(sample);
    model.addEntropy(-1.,varLen);
    model.generate(sample,lambda);
    CHECK(sample.N==lambda,"sample generation failed");
    map.decode(sample);
  }
};

//! GA with pair-wise compression coding
struct PairCodeGA2{
  CodeMap map;
  uint mu,lambda,initL;
  intA E,*init;
  int lo,hi;
  PairCodeGA2(){ mu=30; lambda=100; lo=0; hi=1; init=0; initL=1; }
  void operator()(intAL& sample,std::ostream *evoFile=0){
    if(!sample.N){ if(init) sample.set(lambda,*init); else random(sample,lambda,initL,lo,hi); return; }

    sample.selectBest(mu);
    E=sample(0)->data();//elitist
    sample.weightEqual();

    map.encode(sample,.1);
    sample.resample(lambda);
    cross(sample);
    mutate(sample,lo,MT::MAX(hi,map.maxcode),1.,false);
    //map.encodeAndMute(sample,MutationOp(),lo,hi);
    map.decode(sample);

    sample.add(E);
  }
};


//===========================================================================
//
// The Search class
//

/*! A utility class for search. Evaluation is a struct with and operator()(intA&)
 and Heuristic is a struct with and operator()(intAL&).*/
template<class Evaluation,class Heuristic,class SampleList>
class Search{
public:
  //! the evaluation method (a struct with and operator()(SampleType&) method)
  Evaluation E;

  //! the search heuritic (a struct with and operator()(SampleList&) method)
  Heuristic H;

  //! the sample set (after evaluation typically order by quality)
  SampleList sample;

  //! parameters
  bool hillClimb;
  uint restart,stop;
  uint maxLength;
  
  typedef typename SampleList::node node;

  //! constructor
  Search(){
    hillClimb=false;
    restart=0;
    stop=0;
    maxLength=0;
  }

  /*! iterates (1) generating a search population with the
      heuristic, (2) optionally hillClimbing or cutting the samples,
      (3) evaluating the samples.*/
  void searchStep(node& best,std::ostream *evoFile){
    node n;
    double sampleAvg;

    //generate new samples
    H(sample,evoFile);

    //cut size
    if(maxLength) forall(n,sample) if(n->N>maxLength) n->resizeCopy(maxLength);
    
    //hill climber (and evaluate)
    if(hillClimb) forall(n,sample){ n->w=hillClimber(n->data(),E); if(n->w==1.) break; }
    //if(hillClimb) forall(n,sample){ n->w=gsat(n->data(),E); if(n->w==1.) break; }
    
    //evaluate
    if(!hillClimb) forall(n,sample){ n->w=E(n->data()); if(n->w==1.) break; }
    sample.averageBest(sampleAvg,best);
    
    //report
    if(evoFile) (*evoFile) <<best->w <<' ' <<sampleAvg <<' ' <<E.cost <<std::endl;
    //E(best); //E.display();
  }

  /*! iterates the searchStep function and produces output files; returns the 'best'
      found until time 't' */
  bool search(node& best,uint& t,std::ostream *evoFile=0,std::ostream *genoFile=0){
    sample.clear();

    for(t=0;;t++){
      //restart
      if(restart && t && !(t%restart)){ sample.clear(); }

      if(evoFile) (*evoFile) <<t <<' ';
      searchStep(best,evoFile);

      if(genoFile) (*genoFile) <<t <<' ' <<best->data() <<std::endl;

      //stopping criterion
      if((stop && t>=stop) || (best->w==1.)){
	if(best->w==1.) return true; else return false;
      }
    }
  }

  //! same as above; opens files "z-evo" and "z-gen" as output; reports on std::out
  bool search(){
    node best;
    uint t;
    std::ofstream evofile("z-evo"),genfile("z-gen");
    bool found=search(best,t,&evofile,&genfile);

    if(found){
      std::cout <<"solution found at time " <<t <<", cost=" <<E.cost <<" : " <<best->data() <<std::endl;
      std::cout <<"value=" <<E(best->data()) <<std::endl;
    }else{
      std::cout <<"no solution found! at time " <<t <<", cost=" <<E.cost <<" : " <<best->data() <<std::endl;
      std::cout <<"value=" <<E(best->data()) <<std::endl;
    }

    return found;
  }

  /*! iterates the search 'k' times and collects performance statistics
      which are written to std::out or the given 'statisticsFile' */
  void runs(uint k,std::ostream *statisticsFile=0){
    std::ofstream evofile("z-evo"),runfile("z-run");
    //averaging variables
    double c,tA,tV,eA,eV,bA,bV,f,fc;


    c=tA=tV=eA=eV=bA=bV=f=0.; fc=k;
    node x;
    uint t;
    bool found;

    for(;k--;){
      std::cout <<"** run " <<k <<std::endl;
      E.reset();
      found=search(x,t,&evofile);
      evofile <<std::endl;

      if(found){
	runfile <<t <<' ' <<E.cost <<' ' <<E.Bcost <<' ' <<x->data() <<std::endl;
	std::cout <<"solution found at time " <<t <<", cost=" <<E.cost <<'/'<<E.Bcost <<" : " <<x->data() <<std::endl;
	std::cout <<"value=" <<E(x->data()) <<std::endl;
      }else{
	runfile <<t <<' ' <<E.cost <<' ' <<E.Bcost <<' ' <<x->data() <<" FAILED!" <<std::endl;
	std::cout <<"no solution found! at time " <<t <<", cost=" <<E.cost <<'/'<<E.Bcost <<" : " <<x->data() <<std::endl;
	std::cout <<"value=" <<E(x->data()) <<std::endl;
      }

      if(found){
	f++;
        c+=1.; tA+=t; tV+=(double)t*t; eA+=E.cost; eV+=(double)E.cost*E.cost; bA+=E.Bcost; bV+=(double)E.Bcost*E.Bcost;
      }
    }

    std::cout <<"statistics: time, cost, Bcost (each with standard deviation), found\n"
	<<tA/c <<' ' <<sqrt((tV-tA*tA/c)/c) <<", "
	<<eA/c <<' ' <<sqrt((eV-eA*eA/c)/c) <<", "
	<<bA/c <<' ' <<sqrt((bV-bA*bA/c)/c) <<", "
	<<f/fc <<std::endl;

    if(statisticsFile){
      (*statisticsFile)
	<<tA/c <<' ' <<sqrt((tV-tA*tA/c)/c) <<' '
	<<eA/c <<' ' <<sqrt((eV-eA*eA/c)/c) <<' '
	<<bA/c <<' ' <<sqrt((bV-bA*bA/c)/c) <<' '
	<<f/fc <<' ';
    }
  }
};
//@}
}



using MT::Search;
using MT::RandomSearch;
using MT::PlainGA;
using MT::SplitCodeGA;
using MT::PairCodeGA;
using MT::PairCodeGA2;
using MT::PoolEDA;
using MT::PairCodeEDA;

#endif
