#ifndef _testFunctions_h_
#define _testFunctions_h_

#include "std.h"
#include "array.h"

/*! some test functions */
namespace TestFunctions{
  extern intA kTrapMix; 

  //! the hierarchical XOR function for a string of 0/1s
  double HXOR(byte* s,uint n,bool cleanFromNon01=false);
  
  //! the hierarchical XOR function for a string of 0/1s
  double HXOR(const MT::Array<int>& x,uint cutoff=0,bool cleanFromNon01=false);

  //! counts the number of 1s (or targetSymbols) in a string
  double CountingOnes(const MT::Array<byte>& x,byte targetSymbol=1);

  //! counts the number of 1s (or targetSymbols) in a string
  double CountingOnes(const MT::Array<int>& x,uint cutoff,int targetSymbol=1);

  double kTrap(const intA& x,uint k,uint L);

  inline void sinData(doubleA& X,doubleA& Y,double noise=0.,uint N=1000,double minx=-1,double maxx=1.){
    X.resize(N,1); Y.resize(N,1);
    uint i;
    for(i=0;i<X.d0;i++){
      X(i,0)=minx+(maxx-minx)*i/X.d0;
      Y(i,0)=::sin(5.*X(i,0));
    }
  }
  inline void writeData(char* file,const doubleA& X,const doubleA& Y){
    std::ofstream os;
    MT::open(os,file);
    for(uint i=0;i<X.d0;i++) os <<X[i].ioraw() <<' ' <<Y[i] <<std::endl;
  }

  //! mapping in to out
  struct Linear{
    struct lModel{ doubleA mat,off; }; 
    lModel M;

    Linear(){};
    Linear(uint inDim){ instantiate(inDim,1); };
    Linear(uint inDim,uint outDim){ instantiate(inDim,outDim); };
    
    void instantiate(uint inDim,uint outDim){
      CHECK(inDim && outDim,"dimensions must be non-zero");
      M.off.resize(outDim);       rndUni(M.off,-1.,1.);
      M.mat.resize(outDim,inDim); rndUni(M.mat,-1,1);
    }

    uint inDim(){ return M.mat.d1; }
    uint outDim(){ return M.off.N; }

    void map(doubleA& in,double& out,bool randomInput=false,double noise=0.){
      doubleA o(&out,1);
      map(in,o,randomInput,noise);
    }
    void map(doubleA& in,doubleA& out,bool randomInput=false,double noise=0.){
      CHECK(randomInput || in.N==inDim(),"wrong input dimension");
      if(randomInput){ in.resize(inDim()); rndUni(in,-1,1); }
      out.resize(outDim());
      out = M.off + M.mat*in;
      if(noise) rndGauss(out,noise,true);
    }
  };

  //! mapping in to out, locally linear, discontinuous or smooth switches between
  struct LocalLinearSwitch{
    struct lModel{ doubleA cen,mat,off; double sig; }; 
    MT::Array<lModel> M;
    bool smooth;

    LocalLinearSwitch(){ smooth=false; };
    LocalLinearSwitch(uint inDim,uint nModels){ instantiate(inDim,1,nModels); smooth=false; };
    LocalLinearSwitch(uint inDim,uint outDim,uint nModels){ instantiate(inDim,outDim,nModels); smooth=false; };

    uint inDim(){ return M(0).mat.d1; }
    uint outDim(){ return M(0).off.N; }
    
    void instantiate(uint inDim,uint outDim,uint nModels);

    void map(doubleA& in,double& out,bool randomInput=false,double noise=0.){
      doubleA o(&out,1); map(in,o,randomInput,noise); }
    void map(doubleA& in,doubleA& out,bool randomInput=false,double noise=0.);
  };

  struct MAXSAT{
    MT::Array<int> fct;
    uint vars,clauses;

    //! reads a CNF file
    void readCNF(const char* file);
    //! returns the ratio of satisfied clauses (talk dumps details to std::cout)
    double sat(const MT::Array<byte>& x,std::ostream *talk=0);
    //! same as sat(x)
    double operator()(const MT::Array<byte>& x){ return sat(x); }
    //! displays the problem
    void write(std::ostream& os) const;
  };
  stdOutPipe(MAXSAT);

  inline void smoothData(doubleA& X,doubleA& Y,double noise=0.,uint N=1000,double minx=-1,double maxx=1.){
    grid(X,1,minx,maxx,N);
    LocalLinearSwitch lls(1,1,10); lls.smooth=true;
    lls.map(X,Y,false,noise);
  }




//===========================================================================
//
// collection of standard test functions
//

  struct StdEvaluation{
    bool newProblem;
    uint length,cost,Bcost;
    enum{ none, CountingOnes, HXOR, MAXSAT, PLANTS } problem;
    TestFunctions::MAXSAT ms;
    //TestFunctions::Plants plants;
    
#if 0
    struct hashFunction{
      size_t operator()(const Sequence& x) const{ 
	int k=0;
	for(uint i=0;i<15 && i<x.N;i++) k|=x(i)<<i;
	return k;
      }
      bool operator()(const Sequence& a,const Sequence& b) const{ return a==b; }
    };
    hash_set<Sequence,hashFunction,hashFunction> evaluated;
#endif
    
    StdEvaluation(){
      problem=none;
      reset();
    }
    ~StdEvaluation(){}

    void reset(){
      newProblem=true;
      cost=0;
     }

    //! evaluation function
    double operator()(const intA& x){
      static int msNum=0;
      byteA y;
      double w;
      
      //check the hash set first
#if 0
      Sequence X; (intA&)X=x;
      if(newProblem){
	evaluated.clear();
      }else{
	hash_set<Sequence,hashFunction,hashFunction>::iterator i;
	i=evaluated.find(X);
	if(i!=evaluated.end()){
	  X.w=((Sequence&)(*i)).w;
	  return X.w;
	}
      }
#endif
      
      //increment costs
      cost++;
      Bcost+=x.N;

      //switch to the specific function
      switch(problem){
      case CountingOnes:
	if(newProblem){ newProblem=false; cost=0; Bcost=0; }
	w=TestFunctions::CountingOnes(x,length);
	break;
      case HXOR:
	if(newProblem){ newProblem=false; cost=0; Bcost=0; }
	w=TestFunctions::HXOR(x,length);
	break;
      case MAXSAT:
	if(newProblem){
	  msNum++;
	  msNum=2;
	  ms.readCNF(STRING("../../../libs/SATLIB/uf"<<length<<"/uf"<<length<<"-0"<<msNum<<".cnf"));
	  std::cout <<"opening " <<msNum <<"th MAXSAT problem" <<std::endl;
	  newProblem=false; cost=0; Bcost=0;
	}
	y.copy(x);
	w=ms(y);
	break;
      case PLANTS:
	y.copy(x);
	  //w=plants(y);
	break;
      default: HALT("unknown problem");
      }
      //evaluated.insert(X);
      return w;
    }
  };
}


template<class Function>
void fillIn(doubleA& data,Function& f,uint i=0,uint j=0,uint k=0,double xl=-1.,double xh=1.,double yl=-1.,double yh=1.){
  doubleA fin,fout; fin.resize(f.inDim()); fin=0.;
  if(data.nd==1){
    uint m,M=data.d0;
    for(m=0;m<M;m++){
      fin(i)=xl+(xh-xl)*m/(M-1);
      f.map(fin,fout);
      data(m)=fout(j);
    }
    return;
  }
  if(data.nd==2){
    if(!i && !j && f.inDim()>1) j=1;
    uint m,n,M=data.d0,N=data.d1;
    for(m=0;m<M;m++) for(n=0;n<N;n++){
      fin(j)=yl+(yh-yl)*m/(M-1);
      fin(i)=xl+(xh-xl)*n/(N-1);
      f.map(fin,fout);
      data(m,n)=fout(k);
    }
  }
}


#ifdef MT_IMPLEMENTATION
#  include"testFunctions.cpp"
#endif

//using TestFunctions::Evaluation;

#endif
