#include <cmath>
#include <cstdlib>
#include "algos.h"

#ifdef MT_SHARK
#  define Array MT::Array
#endif

#ifndef max
#  define max(a,b) ( (a)>(b) ? (a) : (b) )
#endif

static double MT_SIGN_SVD(double a,double b){ return b>0 ? fabs(a) : -fabs(a); }

uint MT::svd(const Array<double>& A,
	 Array<double>& U,
	 Array<double>& V,
	 Array<double>& w,bool sort){
  Array<double*> Apointers,Upointers,Vpointers;
  unsigned m = A.d0; /* rows */
  unsigned n = A.d1; /* cols */
  U.resize(m,n);
  V.resize(n,n);
  w.resize(n);
  double **a = A.getPointers(Apointers); /* input matrix */
  double **u = U.getPointers(Upointers); /* left vectors */
  double **v = V.getPointers(Vpointers); /* right vectors */

  int flag;
  unsigned i,its,j,jj,k,l,nm(0),r;
  double anorm,c,f,g,h,s,scale,x,y,z,t;

  Array<double> rv1(n);

  /* copy A to U */
  for(i=0;i<m;i++) for(j=0;j<n;j++) u[i][j] = a[i][j];

  /* householder reduction to bidiagonal form */
  g = scale = anorm = 0.0;

  for(i=0;i<n;i++){
    l = i + 1;
    rv1(i) = scale * g;
    g = s = scale = 0.0;

    if(i<m){
      for(k=i;k<m;k++) scale += fabs(u[k][i]);

      if(scale!=0.0){
	for(k=i;k<m;k++){
	  u[k][i] /= scale;
	  s += u[k][i] * u[k][i];
	}

	f = u[i][i];
	g = -MT_SIGN_SVD(sqrt(s),f);
	h = f * g - s;
	u[i][i] = f - g;

	for(j=l;j<n;j++){
	  s = 0.0;
	  for(k=i;k<m;k++) s += u[k][i] * u[k][j];

	  f = s / h;
	  for(k=i;k<m;k++) u[k][j] += f * u[k][i];
	}

	for(k=i;k<m;k++) u[k][i] *= scale;
      }
    }

    w(i) = scale * g;
    g = s = scale = 0.0;

    if(i<m && i!=n-1){
      for(k=l;k<n;k++)scale += fabs(u[i][k]);

      if(scale!=0.0){
	for(k=l;k<n;k++){
	  u[i][k] /= scale;
	  s += u[i][k] * u[i][k];
	}

	f = u[i][l];
	g = -MT_SIGN_SVD(sqrt(s),f);
	h = f * g - s;
	u[i][l] = f - g;

	for(k=l;k<n;k++) rv1(k) = u[i][k] / h;

	for(j=l;j<m;j++){
	  s = 0.0;
	  for(k=l;k<n;k++) s += u[j][k] * u[i][k];

	  for(k=l;k<n;k++) u[j][k] += s * rv1(k);
	}

	for(k=l;k<n;k++) u[i][k] *= scale;
      }
    }

    anorm = max(anorm,fabs(w(i)) + fabs(rv1(i)));
  }

  /* accumulation of right-hand transformations */
  for(l=i=n;i--;l--){
    if(l<n){
      if(g!=0.0){
	  /* double division avoids possible underflow */
	for(j=l;j<n;j++) v[j][i] = (u[i][j] / u[i][l]) / g;

	for(j=l;j<n;j++){
	  s = 0.0;
	  for(k=l;k<n;k++) s += u[i][k] * v[k][j];

	  for(k=l;k<n;k++) v[k][j] += s * v[k][i];
	}
      }

      for(j=l;j<n;j++) v[i][j] = v[j][i] = 0.0;
    }

    v[i][i] = 1.0;
    g = rv1(i);
  }

  /* accumulation of left-hand transformations */
  for(l=i=(m<n?m:n);i--;l--){
    g = w(i);

    for(j=l;j<n;j++) u[i][j] = 0.0;

    if(g!=0.0){
      g = 1.0 / g;

      for(j=l;j<n;j++){
	s = 0.0;
	for(k=l;k<m;k++) s += u[k][i] * u[k][j];

	/* double division avoids possible underflow */
	f = (s / u[i][i]) * g;

	for(k=i;k<m;k++) u[k][j] += f * u[k][i];
      }

      for(j=i;j<m;j++) u[j][i] *= g;
    }else{
      for(j=i;j<m;j++) u[j][i] = 0.0;
    }

    u[i][i]++;
  }

  /* diagonalization of the bidiagonal form */
  for(k=n;k--;){
    for(its=1;its<=30;its++){
      flag = 1;

      /* test for splitting */
      for(l = k + 1;l--;){
	/* rv1 [0] is always zero, so there is no exit */
	nm = l - 1;

	if(fabs(rv1(l)) + anorm == anorm){
	  flag = 0;
	  break;
	}

	if(fabs(w(nm)) + anorm == anorm) break;
      }

      if(flag){
	/* cancellation of rv1 [l] if l greater than 0 */
	c = 0.0;
	s = 1.0;

	for(i=l;i<=k;i++){
	  f = s * rv1(i);
	  rv1(i) *= c;

	  if(fabs(f) + anorm == anorm) break;

	  g = w(i);
	  h = hypot(f,g);
	  w(i) = h;
	  h = 1.0 / h;
	  c = g * h;
	  s = -f * h;

	  for(j=0;j<m;j++){
	    y = u[j][nm];
	    z = u[j][i];
	    u[j][nm] = y * c + z * s;
	    u[j][i] = z * c - y * s;
	  }
	}
      }

      /* test for convergence */
      z = w(k);

      if(l==k){
	if(z<0.0){
	  w(k) = -z;
	  for(j=0;j<n;j++) v[j][k] = -v[j][k];
	}
	break;
      }

      if(its==30) throw k;

      /* shift from bottom 2 by 2 minor */
      x = w(l);
      nm = k - 1;
      y = w(nm);
      g = rv1(nm);
      h = rv1(k);
      f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2.0 * h * y);
      g = hypot(f,1.0);
      f = ((x - z) * (x + z) + h * ((y / (f + MT_SIGN_SVD(g,f))) - h)) / x;

      /* next qr transformation */
      c = s = 1.0;

      for(j=l;j<k;j++){
	i = j + 1;
	g = rv1(i);
	y = w(i);
	h = s * g;
	g *= c;
	z = hypot(f,h);
	rv1(j) = z;
	c = f / z;
	s = h / z;
	f = x * c + g * s;
	g = g * c - x * s;
	h = y * s;
	y *= c;

	for(jj=0;jj<n;jj++){
	  x = v[jj][j];
	  z = v[jj][i];
	  v[jj][j] = x * c + z * s;
	  v[jj][i] = z * c - x * s;
	}

	z = hypot(f,h);
	w(j) = z;

	/* rotation can be arbitrary if z is zero */
	if(z!=0.0){
	  z = 1.0 / z;
	  c = f * z;
	  s = h * z;
	}

	f = c * g + s * y;
	x = c * y - s * g;

	for(jj=0;jj<m;jj++){
	  y = u[jj][j];
	  z = u[jj][i];
	  u[jj][j] = y * c + z * s;
	  u[jj][i] = z * c - y * s;
	}
      }

      rv1(l) = 0.0;
      rv1(k) = f;
      w(k) = x;
    }
  }

  //sorting:
  if(sort){
    unsigned i,j,k;
    double   p;

    for(i=0;i<n-1;i++){
      p = w(k=i);

      for(j=i+1;j<n;j++) if(w(j)>=p) p = w(k=j);

      if(k!=i){
	w(k) = w(i);
	w(i) = p;

	for(j=0;j<n;j++){
	  p       = v[j][i];
	  v[j][i] = v[j][k];
	  v[j][k] = p;
	}

	for(j=0;j<m;j++){
	  p       = u[j][i];
	  u[j][i] = u[j][k];
	  u[j][k] = p;
	}
      }
    }
  }

  //rank analysis

  for(r=0;r<n && w(r)>0.;r++);

  t = r < n ? fabs(w(n-1)) : 0.0;
  r = 0;
  s = 0.0;
  while(r<n && w(r)>t && w(r)+s>s) s += w(r++);

  return r;
}



uint MT::inverse(const Array<double>& A,Array<double>& inverse){
  unsigned i,j,k,m,n,r;

  m = A.d0;
  n = A.d1;

  Array<double> U,V,w;

  inverse.resize(n,m);

  if(m==0 || n==0) return 0;

  r=svd(A,U,V,w,true);

  for(i=0;i<r;i++) w(i) = 1./w(i);
  for(   ;i<n;i++) w(i) = 0.;

  for(i=0;i<n;i++) for(j=0;j<m;j++){
    double* vi = &V(i,0);
    double* uj = &U(j,0);
    double  t  = 0.;

    for(k=0;k<n;k++) t += vi[k] * w(k) * uj[k];

    inverse(i,j) = t;
  }

  return r;
}

double _determinant(double **A,uint n){
  if(n==1) return A[0][0];
  if(n==2) return A[0][0]*A[1][1]-A[0][1]*A[1][0];
  uint i,j;
  double d=0;
  double **B=new double*[n-1];
  for(i=0;i<n;i++){
    for(j=0;j<n;j++){
      if(j<i) B[j]=&A[j][1];
      if(j>i) B[j-1]=&A[j][1];
    }
    d+=((i&1)?-1.:1.) * A[i][0]*_determinant(B,n-1);
  }
  delete[] B;
  return d;
}

double MT::determinant(const Array<double>& A){
  CHECK(A.nd==2 && A.d0==A.d1,"determinants require a squared 2D matrix");
  Array<double*> B;
  A.getPointers(B);
  return _determinant(B.p,A.d0);
}

#ifndef MT_SHARK
double MT::cofactor(const Array<double>& A,uint i,uint j){
  CHECK(A.nd==2 && A.d0==A.d1,"determinants require a squared 2D matrix");
  Array<double> B=A;
  B.delRow(i);
  B.delColumn(j);
  return ((i&1)^(j&1)?-1.:1) * determinant(B);
}
#endif

void MT::SUS(const Array<double>& p,uint n,Array<uint>& s){
  //following T. Baeck "EA in Theo. and Prac." p120
  s.resize(n);
  double sum=0,ptr=MT::rnd.uni();
  uint i,j=0;
  for(i=0;i<p.N;i++){
    sum+=p(i)*n;
    while(sum>ptr){ s(j)=i; j++; ptr+=1.; }
  }
  //now, 'sum' should = 'n' and 'ptr' has been 'n'-times increased -> 'j=n'
  CHECK(j==n,"error in rnd::SUS(p,n,s) -> p not normalized?");
}

uint MT::SUS(const Array<double>& p){
  double sum=0,ptr=MT::rnd.uni();
  uint i;
  for(i=0;i<p.N;i++){
    sum+=p(i);
    if(sum>ptr) return i;
  }
  HALT("error in rnd::SUS(p) -> p not normalized?");
  return 0;
}

#undef max
#ifdef MT_SHARK
#  undef Array
#endif

//===========================================================================
//
// Linear Statistics
//

void LinearStatistics::compute(){
  if(computed) return;
  CHECK(accum,"no data accumulated in linear statistics");
  MeanX=meanX/accum;
  MeanY=meanY/accum;
  VarX =!(varX  - !(meanX^meanX)/accum)/accum;
  CovXY=!(covXY - !(meanX^meanY)/accum)/accum;
  computed=true;
}

void LinearStatistics::computeZeroMean(){
  if(computed) return;
  CHECK(accum,"no data accumulated in linear statistics");
  MeanX.resize(meanX); MeanX.setZero();
  MeanY.resize(meanY); MeanY.setZero();
  VarX =varX  /(accum*accum);
  CovXY=covXY /(accum*accum);
  computed=true;
}

void LinearStatistics::learn(const doubleA& X,const doubleA& Y,double weight){
  CHECK((X.nd==1 && Y.nd==1) || (X.nd==2 && Y.nd==2),"data must either be batch or online data");
  if(X.nd==1 && Y.nd==1){
    if(!weight) return;
    computed=false;
    if(!accum){ //first data received
      accum=weight;
      meanX=weight*X;
      meanY=weight*Y;
      varX =weight*!(X^X);
      covXY=weight*!(X^Y);
      return;
    }
    CHECK(samedim(X,meanX) && samedim(Y,meanY),"PLS: data dimensions differ");
    uint i,j;
    if(lambda>0.){
      accum=(1.-lambda*weight)*accum + weight;
      meanX=(1.-lambda*weight)*meanX + weight*X;
      meanY=(1.-lambda*weight)*meanY + weight*Y;
      FOR2D(varX,i,j) varX(i,j) =(1.-lambda*weight)*varX(i,j)  + weight*X(i)*X(j);
      covXY=(1.-lambda*weight)*covXY + weight*!(X^Y);
    }else{
      accum+=weight;
      FOR1D(meanX,i  ) meanX(i)  +=weight*X(i);
      FOR1D(meanY,i  ) meanY(i)  +=weight*Y(i);
      FOR2D(varX, i,j)  varX(i,j)+=weight*X(i)*X(j);
      FOR2D(covXY,i,j) covXY(i,j)+=weight*X(i)*Y(j);
    }
  }
  if(X.nd==2 && Y.nd==2){
    CHECK(X.d0==Y.d0,"need same number of in/out samples");
    doubleA ones(X.d0); ones=1.;
    computed=false;
    
    if(!accum){
      accum=X.d0;
      meanX=ones*X;
      meanY=ones*Y;
      covXY=!~X*Y;
      varX =!~X*X;
    }else{
      accum+=X.d0;
      meanX+=ones*X;
      meanY+=ones*Y;
      covXY+=!~X*Y;
      varX +=!~X*X;
    }
  }
}

void LinearStatistics::forget(double lambda){
  computed=false;
  accum=(1.-lambda)*accum;
  meanX=(1.-lambda)*meanX;
  meanY=(1.-lambda)*meanY;
  varX =(1.-lambda)*varX;
  covXY=(1.-lambda)*covXY;
}
void MT::normalizeData(doubleA& X){
  CHECK(X.nd==2,"data has to be a 2D batch");
  uint n,N=X.d0,k,K=X.d1;
  doubleA mean,var,x,sd;
  doubleA ones(N); ones=1.;

  innerProduct(mean,ones,X);
  mean/=(double)N;
  for(n=0;n<N;n++) X[n]()-=mean;

  innerProduct(var,!~X,X);
  var/=(double)N;
  sd.resize(K);
  for(k=0;k<K;k++) sd(k)=sqrt(var(k,k));
  for(n=0;n<N;n++) for(k=0;k<K;k++) if(sd(k)) X(n,k)/=sd(k);
}

doubleA XSpline::eval(double t){
  CHECK(t>=0 && t<V.d0-.99,"out of range");
  if(!W.N){
    W.resize(V.d0); W=1.; W(0)=W(W.N-1)=0.;
  }
  // First we should find our parameter interval
  // 
  // In accordance with the article, this function defines the value
  // of the spline in the range [ tk+1, tk+2 ], so k is ...
  double kf = floor(t) - 1.0;
  int k = int(kf);
  
  double W0,W1,W2,W3;
  if(k>=0 && k<(int)W.N)    W0=W(k);   else W0=0.;
  if(k+1<(int)W.N) W1=W(k+1); else W1=0.;
  if(k+2<(int)W.N) W2=W(k+2); else W2=0.;
  if(k+3<(int)W.N) W3=W(k+3); else W3=0.;
  
  // Find the big T values
  // 
  // T0p is the point where fk decreases to 0
  // T1p is the point where fk+1 decreases to 0
  // T2m is the point where fk+2 increases from 0
  // T3m is the point where fk+3 increases from 0
  //
  // The parameter sk of a control point k affects
  // the functions that have max vals in k-1 and k+1
  // Therefore, sv[k+1] affects T0p (The point where
  // fk decreases to 0) and T2m. Similarly, sv[k+2]
  // affects T1p and T3m
  double T0p = kf+1 + (W1>0?W1:0) * DELTA;
  double T1p = kf+2 + (W2>0?W2:0) * DELTA;
  double T2m = kf+1 - (W1>0?W1:0) * DELTA;
  double T3m = kf+2 - (W2>0?W2:0) * DELTA;
  
  // Find the pk values
  // The pm values are derived directly from the Ts above
  double tmp_factor = 2/(DELTA*DELTA);
  double pm1 = intpow(kf   - T0p, 2) * tmp_factor; 
  double pm0 = intpow(kf+1 - T1p, 2) * tmp_factor; 
  double pp1 = intpow(kf+2 - T2m, 2) * tmp_factor; 
  double pp2 = intpow(kf+3 - T3m, 2) * tmp_factor; 
  
  
  // The sv indices are determined like for the Ts
  double qp0 = W1<0?-W1/2.0:0;
  double qp1 = W2<0?-W2/2.0:0;
  double qp2 = W1<0?-W1/2.0:0;
  double qp3 = W2<0?-W2/2.0:0;
  
  
  // The code below resembles the part of (17) in the article, 
  // where A0 ... A3 are calculated, but there are differences
  // 
  // 1.
  // 
  // We use g instead of f. This is ok, because 
  //    q = 0 => g(u) = f(u) 
  //
  // 2.
  // we don't check whether we are inside the legal range 
  // of fk+1 and fk+2. This is not necessary, since we are in the
  // range [ tk+1, tk+2 ] and both gk+1 and gk+2 must be defined 
  // in this range.
  // 
  // 3.
  //
  // if we are in the ranges where fk and fk+3 would have been
  // nil in the case of basic splines, we use h instead of g
  // except if q
  //
  double A0,A1,A2,A3;      
  if (t <= T0p)
    A0 =  g((t-T0p)/(kf   - T0p), qp0, pm1);
  else
    A0 = qp0>0 ? h((t-T0p)/(kf   - T0p), qp0) : 0;
  
  A1 = g((t-T1p)/(kf+1 - T1p), qp1, pm0);
  
  A2 = g((t-T2m)/(kf+2 - T2m), qp2, pp1);
  
  if (t >= T3m)
    A3 = g((t-T3m)/(kf+3 - T3m), qp3, pp2);
  else
    A3 = qp3>0 ? h((t-T3m)/(kf+3 - T3m), qp3) : 0;
  
  double SUM = A0+A1+A2+A3;
  
  doubleA V0,V1,V2,V3;
  if(k>=0 && k<(int)V.d0)     V0.referToSubDim(V,k);   else V0.referToSubDim(V,0);
  if(k+1<(int)V.d0) V1.referToSubDim(V,k+1); else V1.referToSubDim(V,V.d0-1);
  if(k+2<(int)V.d0) V2.referToSubDim(V,k+2); else V2.referToSubDim(V,V.d0-1);
  if(k+3<(int)V.d0) V3.referToSubDim(V,k+3); else V3.referToSubDim(V,V.d0-1);
  
  doubleA C =  V0*A0+V1*A1+V2*A2+V3*A3;
  C /= SUM; // Normalise
  return C;
}
